/*
 * File: hash_table.c
 * ------------------
 * Hash table.
 *
 * By: Marcus 2018
 */

#include "hash_table.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#ifdef HT_REF_COUNT
static int sTables = 0;
#endif

/*
 * Function: HT_Create
 * -------------------
 */
HashTable *HT_Create(unsigned int capacity) {
    HashTable *ht;
    int i;

    ht = (HashTable *)malloc(sizeof(HashTable));
    if (!capacity) capacity = HT_DEF_CAPACITY;
  
    ht->list = (HashEntry **)malloc(sizeof(HashEntry *)*capacity);
    for (i = 0; i < capacity; i++) ht->list[i] = 0;
  
    ht->capacity = capacity;
    ht->minCapacity = capacity;
    ht->entries = 0;

#ifdef HT_REF_COUNT
    sTables++;
#endif

    return ht;
}

/*
 * Function: HT_Clear
 * ------------------
 */
void HT_Clear(HashTable *ht, void (*delFunc)(void *)) {
    int i;
    if (!ht || !ht->list) return;
    for (i = 0; i < ht->capacity; i++) {
        if (ht->list[i]) {
            HashEntry *e = ht->list[i];
            while (e) {
                HashEntry *next = e->next;
                if (e->data && delFunc) delFunc(e->data);
                free(e->key);
                free(e);
                e = next;
                ht->entries--;
            }
            ht->list[i] = 0;
        }
    }
    assert(ht->entries == 0);
}

/*
 * Function: HT_Free
 * -----------------
 */
void HT_Free(HashTable *ht, void (*delFunc)(void *)) {
    if (ht) {
        if (ht->list) {
            HT_Clear(ht, delFunc);
            free(ht->list);
        }
        free(ht);
#ifdef HT_REF_COUNT
        sTables--;
#endif
    }
}

/*
 * Function: ApplyDataFunction
 * ---------------------------
 */
void HT_ApplyDataFunction(HashTable *ht, void (*func)(void *)) {
    int i;
    if (!ht || !ht->list) return;
    for (i = 0; i < ht->capacity; i++) {
        if (ht->list[i]) {
            HashEntry *e = ht->list[i];
            while (e) {
                if (e->data) func(e->data);
                e = e->next;
            }
        }
    }
}

void HT_ApplyKeyFunction(HashTable *ht, void (*func)(char *, void *)) {
    int i;
    if (!ht || !ht->list) return;
    for (i = 0; i < ht->capacity; i++) {
        if (ht->list[i]) {
            HashEntry *e = ht->list[i];
            while (e) {
                func(e->key, e->data);
                e = e->next;
            }
        }
    }
}

/*
 * Function: HT_Hash
 * -----------------
 */
unsigned int HT_Hash(HashTable *ht, const char *key) {
    unsigned int hash = 5381;
    int c;

    while ((c = *key++)) hash = ((hash << 5) + hash) + c;

    return hash%ht->capacity;
}

/*
 * Function: HT_Add
 * ----------------
 */
int HT_Add(HashTable *ht, const char *key, void *data) {
    unsigned int hash;
    HashEntry *e;

    if (!ht) return 0;
    
    if ((double)ht->entries/(double)ht->capacity > HT_GROW_AT) {
        HT_Resize(ht, ht->capacity*2);
    }

    hash = HT_Hash(ht, key);

    if (ht->list[hash] == 0) {
        e = (HashEntry *)malloc(sizeof(HashEntry));
        ht->list[hash] = e;
    }
    else {
        e = ht->list[hash];
        while (e) {
            if (strcmp(key, e->key) == 0) {
                return 0;
            }
            if (!e->next) {
                e->next = (HashEntry *)malloc(sizeof(HashEntry));
                e = e->next;
                break;
            }
            e = e->next;
        }
    }
    e->key = strdup(key);
    e->next = 0;
    e->data = data;
    ht->entries++;

    return 1;
}

void HT_UnsafeAdd(HashTable *ht, const char *key, void *data) {
    unsigned int hash;
    HashEntry *e;

    if (!ht) return;
    
    if ((double)ht->entries/(double)ht->capacity > HT_GROW_AT) {
        HT_Resize(ht, ht->capacity*2);
    }

    e = (HashEntry *)malloc(sizeof(HashEntry));
    e->key = strdup(key);
    e->data = data;

    hash = HT_Hash(ht, key);

    if (ht->list[hash] == 0) {
        e->next = 0;
        ht->list[hash] = e;
    }
    else {
        e->next = ht->list[hash];
        ht->list[hash] = e;
    }

    ht->entries++;
}

/*
 * Function: HT_AddEntry
 * ---------------------
 * Used when resizing.
 */
void HT_AddEntry(HashTable *ht, HashEntry *e) {
    unsigned int hash;

    if (!ht) return;
    
    hash = HT_Hash(ht, e->key);

    if (ht->list[hash] == 0) {
        e->next = 0;
        ht->list[hash] = e;
    }
    else {
        e->next = ht->list[hash];
        ht->list[hash] = e;
    }

    ht->entries++;

}


/*
 * Function: HT_Delete
 * -------------------
 */
int HT_Delete(HashTable *ht, const char *key, void (*delFunc)(void *)) {
    unsigned int hash;
    if (!ht) return 0;

    if (ht->capacity/2 >= ht->minCapacity &&
        (double)ht->entries/(double)ht->capacity < HT_SHRINK_AT) {
        HT_Resize(ht, ht->capacity/2);
    }

    hash = HT_Hash(ht, key);

    if (!ht->list[hash]) return 0;

    HashEntry *e = ht->list[hash];
    if (strcmp(e->key, key) == 0) {
        HashEntry *next = e->next;
        if (delFunc) delFunc(e->data);
		free(e->key);
        free(e);
        ht->list[hash] = next;
        ht->entries--;
        return 1;
    }
    else {
        for (; e->next; e = e->next) {
            if (strcmp(e->next->key, key) == 0) {
                HashEntry *next = e->next->next;
                if (delFunc) delFunc(e->next->data);
				free(e->next->key);
                free(e->next);
                e->next = next;
                ht->entries--;
                return 1;
            }
        }
    }
    return 0;
}

/*
 * Function: HT_Function
 * ---------------------
 */
int HT_Function(HashTable *ht, const char *key, void (*func)(void *)) {
    unsigned int hash;
    if (!ht) return 0;

    hash = HT_Hash(ht, key);

    if (!ht->list[hash]) return 0;

    HashEntry *e = ht->list[hash];
    if (strcmp(e->key, key) == 0) {
        func(e->data);
        return 1;
    }
    else {
        for (; e->next; e = e->next) {
            if (strcmp(e->next->key, key) == 0) {
                func(e->next->data);
                return 1;
            }
        }
    }
    return 0;
}

/*
 * Function: HT_Get
 * ----------------
 */
void *HT_Get(HashTable *ht, const char *key) {
    unsigned int hash;
    HashEntry *e;

    if (!ht || !ht->list) return 0;

    hash = HT_Hash(ht, key);
    for (e = ht->list[hash]; e  != 0; e = e->next)
        if (strcmp(e->key, key) == 0) return e->data;

    return 0;
}

/*
 * Function: HT_GetEntries
 * -----------------------
 * Get a list of all entries in table.
 */
HashEntry *HT_GetEntries(HashTable *ht) {
    HashEntry *list = 0;
    HashEntry *e = 0;
    int i;
    
    if (!ht || !ht->list) return 0;
    for (i = 0; i < ht->capacity; i++) {
        if (ht->list[i]) {
            if (e) {
                e->next = (HashEntry *)malloc(sizeof(HashEntry));
                e->next->data = ht->list[i]->data;
                e->next->key = ht->list[i]->key;
                e->next->next = 0;
                e = e->next;
            }
            else {
                list = (HashEntry *)malloc(sizeof(HashEntry));
                list->data = ht->list[i]->data;
                list->key = ht->list[i]->key;
                list->next = 0;
                e = list;
            }
        }
    }
    return list;
}

/* SLOW. */
int HT_GetKey(HashTable *ht, const void *data, char *key) {
    int i;
    HashEntry *e;

    if (!ht || !ht->list || !data) return 0;
    for (i = 0; i < ht->capacity; i++) {
        for (e = ht->list[i]; e; e = e->next) {
            if (e->data == data) {
                strcpy(key, e->key);
                return 1;
            }
        }
    }
    return 0;
}

/*
 * Function: HT_Exists
 * -------------------
 */
int HT_Exists(HashTable *ht, const char *key) {
	return HT_Get(ht, key) != 0;
}

/*
 * Function: HT_Resize
 * -----------------------
 */
void HT_Resize(HashTable *ht, unsigned int capacity) {
	HashEntry **oldList;
	unsigned int oldCapacity;
	int i;

	if (!ht) return;
    if (capacity < 1) capacity = 1;
    if (ht->minCapacity > capacity) ht->minCapacity = capacity;
	if (ht->capacity == capacity) return;
	
	oldList = ht->list;
	oldCapacity = ht->capacity;

	ht->list = (HashEntry **)malloc(sizeof(HashEntry *)*capacity);
	ht->capacity = capacity;
	for (i = 0; i < capacity; i++) ht->list[i] = 0;
	
	for (i = 0; i < oldCapacity; i++) {
		HashEntry *e = oldList[i];
		while (e) {
			/*HashEntry *next = e->next;
			HT_Add(ht, e->key, e->data);
			free(e->key);
			free(e);
			ht->entries--;
			e = next;*/

            /* Keep old entry. */
            HashEntry *next = e->next;
            HT_AddEntry(ht, e);
            ht->entries--;
            e = next;
		}
	}
	free(oldList);
}

int HT_TableCount() {
#ifdef HT_REF_COUNT
    return sTables;
#else
    return 0;
#endif
}

