/*
 * File: memory_manager.c
 * ----------------------
 * Prolan memory manager.
 *
 * By: Marcus 2018
 */

#include "memory_manager.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

/* Base hash table. */
HashTable *sTable = 0;

/* Used while building list of keys. */
static KeyList *sKeyList = 0;

/*
 * Function: Var__AddReference
 * ---------------------------
 */
void Var__AddReference(Variable *var, Variable *ref);

/*
 * Function: Var__RemoveReference
 * ------------------------------
 */
void Var__RemoveReference(Variable *var, Variable *ref);

/*
 * Function: DeleteVar
 * -------------------
 * Delete variable recursively, used with HT_Delete.
 *
 * data: variable
 */
void DeleteVar(void *data) {
    Variable *var = (Variable *)data;
    if (!var) return;
    if (var->ht) {
        HT_Free(var->ht, DeleteVar);
    }
    if (var->type == VAR_STRING) free(var->value.s);
    while (var->refs) {
        Reference *next = var->refs->next;
        free(var->refs);
        var->refs = next;
    }
    free(var);
}

/*
 * Function: UnreferenceVar
 * ------------------------
 * Remove reference connections recursively, used with HT_Function.
 *
 * data: variable
 */
void UnreferenceVar(void *data) {
    Variable *var = (Variable *)data;
    if (!var) return;
    if (var->type == VAR_ALIAS) {
        Var__RemoveReference(var->value.a, var);
    }
    Reference *r = var->refs;
    while (r) {
        r->var->type = VAR_UNDEFINED;
        r->var->value.a = 0;
        r = r->next;
    }
    if (var->ht) HT_ApplyDataFunction(var->ht, UnreferenceVar);
}

/*
 * Function: CollectKeys
 * ---------------------
 * Used for building a list of keys, used with HT_ApplyKeyFunction.
 *
 * key: key to variable
 * data: the variable, not used.
 */
void CollectKeys(char *key, void *data) {
    KeyList *k = (KeyList *)malloc(sizeof(KeyList));
    k->key = key;
    k->next = 0;

    if (sKeyList == 0) {
        sKeyList = k;
    }
    else {
        k->next = sKeyList;
        sKeyList = k;
    }
}

/*
 * Function: MM_Init
 * -----------------
 * Init static memory manager.
 */
void MM_Init() {
    if (sTable) MM_Terminate();
    sTable = HT_Create(16);
}

/*
 * Function: MM_Free
 * -----------------
 * Free static memory manager and all allocated variables.
 */
void MM_Terminate() {
    if (sTable) {
        HT_Free(sTable, DeleteVar);
        sTable = 0;
    }
    assert(HT_TableCount() == 0);
}

/*
 * Function: MM_GetVar
 * -------------------
 * Return a Variable if it exists.
 *
 * name: name of the variable
 *
 * Returns: a Variable or 0
 */
Variable *MM_GetVar(const char *name) {
    HashTable *ht = sTable;
    Variable *var;
    char tmp[64];
    int i;
    
    do {
        i = 0;
        while (*name && *name != '.') tmp[i++] = *name++;
        tmp[i] = '\0';
        var = HT_Get(ht, tmp);
        if (!var) return 0;
        if (var->type == VAR_ALIAS) {
            var = var->value.a;
        }
        if (*name) {
            if (!var->ht) return 0;
            ht = var->ht;
            name++;
        }
    } while (*name);

    return var;
}

/*
 * Function: MM_GetKeys
 * --------------------
 * Return all subkeys of a variable. The list needs to be freed by the caller
 * using MM_FreeKeys.
 *
 * name: name of the variable.
 *
 * Returns: Linked list of keys.
 */
KeyList *MM_GetKeys(const char *name) {
    Variable *var;

    var = MM_GetVar(name);
    if (!var || !var->ht) return 0;

    sKeyList = 0;    
    HT_ApplyKeyFunction(var->ht, CollectKeys);

    return sKeyList;
}

/*
 * Function: MM_FreeKeys
 * ---------------------
 * Free list of keys.
 *
 * list: keys
 */
void MM_FreeKeys(KeyList *list) {
    while (list) {
        KeyList *next = list->next;
        free(list);
        list = next;
    }
}

/*
 * Function: MM_CreateVar
 * ----------------------
 * Return a new or existing Variable.
 *
 * name: name of the variable
 *
 * Returns: a Variable
 */
Variable *MM_CreateVar(const char *name) {
    /*HashTable *ht = sTable;
    Variable *var;
    char tmp[64];
    int i;
    
    do {
        i = 0;
        while (*name && *name != '.') tmp[i++] = *name++;
        tmp[i] = '\0';
        var = HT_Get(ht, tmp);
        if (!var) {
            var = (Variable *)malloc(sizeof(Variable));
            var->type = VAR_UNDEFINED;
            var->ht = 0;
            var->refs = 0;
            var->owner = ht;
            HT_UnsafeAdd(ht, tmp, var);
        }
        else if (var->type == VAR_ALIAS) {
            var = var->value.a;
            if (parent) *parent = var;
        }
        if (*name) {
            if (parent) *parent = var;
            if (!var->ht) var->ht = HT_Create(4);
            ht = var->ht;
            name++;
        }
    } while (*name);

    return var;*/

    HashTable *ht = sTable;
    Variable *parent = 0;
    Variable *var;
    char tmp[64];
    int i;
    
    do {
        i = 0;
        while (*name && *name != '.') tmp[i++] = *name++;
        tmp[i] = '\0';
        var = HT_Get(ht, tmp);
        if (!var) {
            var = (Variable *)malloc(sizeof(Variable));
            var->type = VAR_UNDEFINED;
            var->ht = 0;
            var->refs = 0;
            var->parent = parent;
            HT_UnsafeAdd(ht, tmp, var);
        }
        else if (var->type == VAR_ALIAS) {
            var = var->value.a;
        }
        if (*name) {
            if (!var->ht) var->ht = HT_Create(4);
            ht = var->ht;
            parent = var;
            name++;
        }
    } while (*name);

    return var;
}

/*
 * Function: MM_DeleteVar
 * ----------------------
 * Delete a variable and all its sub variables.
 *
 * name: name of the variable
 */
void MM_DeleteVar(const char *name) {
    HashTable *ht = sTable;
    Variable *var;
    char tmp[64];
    int i;
    
    do {
        i = 0;
        while (*name && *name != '.') tmp[i++] = *name++;
        tmp[i] = '\0';
        var = HT_Get(ht, tmp);
        if (!var) return;
        /* Note that tmp is wrong here! Deleting a reference will delete the
           reference variable - not the one it's pointing at. The referenced
           variable may reside in ANY hash table, so we can't easily get its
           name ... Should a variable know about its table? 
               Ok, fixed that, added an owner to Variable and a function to get
           the key of a variable in a hash table (this is slow, have to parse
           every entry in the table while looking for the variable.
               I'm not even sure if we want this behavior. */
        /* if (var->type == VAR_ALIAS) var = var->value.a; */
        if (var->type == VAR_ALIAS) {
            var = var->value.a;
            if (var->parent) {
                ht = var->parent->ht;
            }
            else {
                ht = sTable;
            }
            //ht = var->owner;
            HT_GetKey(ht, var, tmp);
        }
        if (*name) {
            if (!var->ht) return;
            ht = var->ht;
            name++;
        }
    } while (*name);
    /* Have to deal with references first to avoid memory mess. */
    HT_Function(ht, tmp, UnreferenceVar);
    HT_Delete(ht, tmp, DeleteVar);
}

/*
 * Function: MM_UnaliasVar
 * -----------------------
 * Unalias variable.
 *
 * name: name of the variable.
 */
void MM_UnaliasVar(const char *name) {
    HashTable *ht = sTable;
    Variable *var;
    char tmp[64];
    int i;
    
    do {
        i = 0;
        while (*name && *name != '.') tmp[i++] = *name++;
        tmp[i] = '\0';
        var = HT_Get(ht, tmp);
        if (!var) return;
        if (*name) {
            if (var->type == VAR_ALIAS) var = var->value.a;
            if (!var->ht) return;
            ht = var->ht;
            name++;
        }
    } while (*name);
    if (var->type == VAR_ALIAS) {
        Var_Free(var);
    }
}


/*
 * Function: MM_CopyVar
 * ----------------------
 * Copy a variable and all its sub variables.
 *
 * dst: destination variable
 * src: source variable
 */
void MM_CopyVar(Variable *dst, Variable *src) {
    Var_Set(dst, src);
    if (!src) return;
    if (src->ht) {
        int i;
        dst->ht = HT_Create(src->ht->capacity);
        for (i = 0; i < src->ht->capacity; i++) {
            HashEntry *e = src->ht->list[i];
            while (e) {
                Variable *newVar;
                HashEntry *dste = (HashEntry *)malloc(sizeof(HashEntry));
                dste->key = strdup(e->key);
                newVar = (Variable *)malloc(sizeof(Variable));
                newVar->type = VAR_UNDEFINED;
                newVar->parent = dst;
                newVar->refs = 0;
                newVar->ht = 0;
                dste->data = newVar;
                MM_CopyVar(newVar, e->data);
                dste->next = dst->ht->list[i];
                dst->ht->list[i] = dste;
                dst->ht->entries++;
                e = e->next;
            }
        }
    }
}

/*
 * Function: MM_VarDefined
 * -----------------------
 * Returns variable exists and is not undefined.
 *
 * name: name of the variable
 *
 * Returns: 1 or 0    
 */
int MM_VarDefined(const char *name) {
    HashTable *ht = sTable;
    Variable *var;
    char tmp[64];
    int i;
    
    do {
        i = 0;
        while (*name && *name != '.') tmp[i++] = *name++;
        tmp[i] = '\0';
        var = HT_Get(ht, tmp);
        if (!var) return 0;
        if (var->type == VAR_ALIAS) var = var->value.a;
        if (*name) {
            if (!var->ht) return 0;
            ht = var->ht;
            name++;
        }
    } while (*name);

    /*return var != 0;*/
    return var && var->type != VAR_UNDEFINED;
}

/*
 * Function: MM_VarSubCount
 * ------------------------
 * Return number of sub variables for a variable.
 *
 * name: name of the variable
 *
 * Returns: number of sub variables
 */
int MM_VarSubCount(const char *name) {
    Variable *var = MM_GetVar(name);
    if (!(var && var->ht)) return 0;
    return var->ht->entries;
}

/*
 * Function: Var_New()
 * -------------------
 * Return new undefined variable.
 *
 * Returns: undefined variable.
 */
Variable *Var_New() {
    Variable *var = (Variable *)malloc(sizeof(Variable));
    var->type = VAR_UNDEFINED;
    var->parent = 0;
    var->ht = 0;
    var->refs = 0;
    return var;
}


/*
 * Function: Var_SetNumber
 * --------------------
 * Assign a numeric value to a variable.
 *
 * var:   variable
 * value: number
 */
void Var_SetNumber(Variable *var, double value) {
    if (var->type == VAR_STRING) free(var->value.s);
    var->type = VAR_NUMBER;
    var->value.d = value;
}

/*
 * Function: Var_GetNumber
 * -----------------------
 * Get numeric value of variable.
 *
 * var: numeric variable
 *
 * Returns: number
 */
double Var_GetNumber(Variable *var) {
    assert(var->type == VAR_NUMBER);
    return var->value.d;
}

/*
 * Function: Var_SetString
 * -----------------------
 * Assign a string value to a variable.
 *
 * var:   variable
 * value: string
 */
void Var_SetString(Variable *var, const char *value) {
    if (var->type == VAR_STRING) free(var->value.s);
    var->type = VAR_STRING;
    var->value.s = strdup(value);
}

/*
 * Function: Var_GetString
 * -----------------------
 * Get string value of variable.
 *
 * var: string variable
 *
 * Returns: string
 */
const char *Var_GetString(Variable *var) {
    assert(var->type == VAR_STRING);
    return var->value.s;
}

/*
 * Function: Var_SetFunction
 * -------------------------
 * Assign function index, to variable.
 *
 * var: variable
 * func: function index
 */
void Var_SetFunction(Variable *var, int func) {
    if (var->type == VAR_STRING) free(var->value.s);
    var->type = VAR_FUNCTION;
    var->value.i = func;
}

/*
 * Function: Var_SetAlias
 * ----------------------
 * Make a variable an alias for another variable.
 *
 * var:   variable that should become an alias
 * value: name of the aliased variable
 */
void Var_SetAlias(Variable *var, const char *value) {
    Variable *src;
    Var_Free(var);
    
    if ((src = MM_GetVar(value))) {
        if (src == var) return;
        var->type = VAR_ALIAS;
        var->value.a = src;
        Var__AddReference(src, var);
    }
}

/*
 * Function: Var__AddReference
 * ---------------------------
 */
void Var__AddReference(Variable *var, Variable *ref) {
    Reference *newRef = (Reference *)malloc(sizeof(Reference));
    newRef->var = ref;
    newRef->next = 0;
    if (var->refs) {
        Reference *r = var->refs;
        while (r->next) r = r->next;
        r->next = newRef;
    }
    else {
        var->refs = newRef;
    }
}

/*
 * Function: Var__RemoveReference
 * ------------------------------
 */
void Var__RemoveReference(Variable *var, Variable *ref) {
    if (!var->refs) return;

    if (var->refs->var == ref) {
        Reference *next = var->refs->next;
        free(var->refs);
        var->refs = next;
    }
    else {
        Reference *r = var->refs;
        while (r->next) {
            if (r->next->var == ref) {
                Reference *next = r->next->next;
                free(r->next);
                r->next = next;
                break;
            }
            r = r->next;
        }
    }
}

/*
 * Function: Var_GetAlias
 * ----------------------
 * Return the name of an aliased variable.
 *
 * var: variable
 *
 * Returns: name of the aliased variable
 */
Variable *Var_GetAlias(Variable *var) {
    assert(var->type == VAR_ALIAS);
    return var->value.a;
}

/*
 * Function: Var_Set
 * -----------------
 * Copy value from one variable to another.
 *
 * dst: destination variable
 * src: source variable
 */
void Var_Set(Variable *dst, Variable *src) {
    Var_Free(dst);
    if (src) {
        dst->type = src->type;
        switch (src->type) {
        case VAR_NUMBER:
            dst->value.d = src->value.d;
            break;
        case VAR_STRING:
            dst->value.s = strdup(src->value.s);
            break;
        case VAR_ALIAS:
            dst->value.a = src->value.a;
            Var__AddReference(src->value.a, dst);
            break;
        case VAR_FUNCTION:
            dst->value.i = src->value.i;
            break;
        default:
            break;
        }
    }
}

/*
 * Function: Var_Free
 * ------------------
 * Make variable undefined.
 *
 * var: variable
 */
void Var_Free(Variable *var) {
    if (var->type == VAR_STRING) {
        free(var->value.s);
    }
    else if (var->type == VAR_ALIAS) {
        Var__RemoveReference(var->value.a, var);
    }
    /* No, we should not look at this variable's refs, because it's not
       being DELETED here. Other variables may still be references to this. */
    var->type = VAR_UNDEFINED;
}


