/*
 * File: memory_manager.h
 * ----------------------
 * Prolan memory manager.
 *
 * By: Marcus 2018
 */

#ifndef __MEMORY_MANAGER_H__
#define __MEMORY_MANAGER_H__

#include "hash_table.h"

/*
 * Enum: VarType
 * -------------
 */
typedef enum {
    VAR_UNDEFINED = 0,
    VAR_NUMBER,
    VAR_STRING,
    VAR_ALIAS,
    VAR_FUNCTION
} VarType;

typedef struct sVariable Variable;
typedef struct sReference Reference;

/*
 * Struct: Reference
 * -----------------
 */
struct sReference {
    Variable *var;
    Reference *next;
};

/*
 * Struct: Variable
 * ----------------
 */
struct sVariable {
    VarType type;
    union {
        char *s;     /* VAR_STRING. */
        double d;    /* VAR_NUMBER. */
        int i;       /* VAR_FUNCTION, bool */
        Variable *a; /* VAR_ALIAS. */
    } value;
    Variable *parent;
    HashTable *ht;
    Reference *refs;
};

/*
 * Struct: KeyList
 * ---------------
 */
typedef struct sKeyList KeyList;
struct sKeyList {
    const char *key;
    KeyList *next;
};

/*
 * Function: MM_Init
 * -----------------
 * Init static memory manager.
 */
void MM_Init();

/*
 * Function: MM_Free
 * -----------------
 * Free static memory manager and all allocated variables.
 */
void MM_Terminate();

/*
 * Function: MM_GetVar
 * -------------------
 * Return a Variable if it exists.
 *
 * name: name of the variable
 *
 * Returns: a Variable or 0
 */
Variable *MM_GetVar(const char *name);

/*
 * Function: MM_GetKeys
 * --------------------
 * Return all subkeys of a variable. The list needs to be freed by the caller
 * using MM_FreeKeys.
 *
 * name: name of the variable.
 *
 * Returns: Linked list of keys.
 */
KeyList *MM_GetKeys(const char *name);

/*
 * Function: MM_FreeKeys
 * ---------------------
 * Free list of keys.
 *
 * list: keys
 */
void MM_FreeKeys(KeyList *list);

/*
 * Function: MM_CreateVar
 * ----------------------
 * Return a new or existing Variable.
 *
 * name: name of the variable
 *
 * Returns: a Variable
 */
Variable *MM_CreateVar(const char *name);

/*
 * Function: MM_DeleteVar
 * ----------------------
 * Delete a variable and all its sub variables.
 *
 * name: name of the variable
 */
void MM_DeleteVar(const char *name);

/*
 * Function: MM_UnaliasVar
 * -----------------------
 * Unalias variable.
 *
 * name: name of the variable.
 */
void MM_UnaliasVar(const char *name);

/*
 * Function: MM_CopyVar
 * ----------------------
 * Copy a variable and all its sub variables.
 *
 * dst: destination variable
 * src: source variable
 */
void MM_CopyVar(Variable *dst, Variable *src);

/*
 * Function: MM_VarDefined
 * -----------------------
 * Returns 1 if variable exists and is not undefined.
 *
 * name: name of the variable
 *
 * Returns: 1 or 0    
 */
int MM_VarDefined(const char *name);

/*
 * Function: MM_VarSubCount
 * ------------------------
 * Return number of sub variables for a variable.
 *
 * name: name of the variable
 *
 * Returns: number of sub variables
 */
int MM_VarSubCount(const char *name);


/*
 * Function: Var_New()
 * -------------------
 * Return new undefined variable.
 *
 * Returns: undefined variable.
 */
Variable *Var_New();

/*
 * Function: Var_SetNumber
 * --------------------
 * Assign a numeric value to a variable.
 *
 * var:   variable
 * value: number
 */
void Var_SetNumber(Variable *var, double value);

/*
 * Function: Var_GetNumber
 * -----------------------
 * Get numeric value of variable.
 *
 * var: numeric variable
 *
 * Returns: number
 */
double Var_GetNumber(Variable *var);

/*
 * Function: Var_SetString
 * -----------------------
 * Assign a string value to a variable.
 *
 * var:   variable
 * value: string
 */
void Var_SetString(Variable *var, const char *value);

/*
 * Function: Var_GetString
 * -----------------------
 * Get string value of variable.
 *
 * var: variable
 *
 * Returns: string
 */
const char *Var_GetString(Variable *var);

/*
 * Function: Var_SetFunction
 * -------------------------
 * Assign function index to variable.
 *
 * var: variable
 * func: function index
 */
void Var_SetFunction(Variable *var, int func);

/*
 * Function: Var_SetAlias
 * ----------------------
 * Make a variable an alias for another variable.
 *
 * var:   variable that should become an alias
 * value: name of the aliased variable
 */
void Var_SetAlias(Variable *var, const char *value);

/*
 * Function: Var_GetAlias
 * ----------------------
 * Return the name of an aliased variable.
 *
 * var: variable
 *
 * Returns: name of the aliased variable
 */
Variable *Var_GetAlias(Variable *var);

/*
 * Function: Var_Set
 * -----------------
 * Copy value from one variable to another.
 *
 * dst: destination variable
 * src: source variable
 */
void Var_Set(Variable *dst, Variable *src);

/*
 * Function: Var_Free
 * ------------------
 * Make variable undefined.
 *
 * var: variable
 */
void Var_Free(Variable *var);

/*
 * Function: Var__AddReference
 * ---------------------------
 */
void Var__AddReference(Variable *var, Variable *ref);

/*
 * Function: Var__RemoveReference
 * ------------------------------
 * Uhm, this one should be internal, but I needed direct access for doing
 * selfish business in function calls.
 */
void Var__RemoveReference(Variable *var, Variable *ref);

#endif

