/*
 * File: prolan.c
 * --------------
 * Prolan interpreter.
 *
 * By: Marcus 2018
 *
 * Fixes:
 * * SkipBlock should store a destination token the first time and re-use it in
 *   future skips. This will improve execution time.
 * * To speed hash tables up, ALWAYS move an element to the head of the table
 *   when being accessed. On the other hand ... this could probably make
 *   repeated loops slower, since the first element accessed on the first run
 *   will be last bucket on the second run. Hrm. Maybe just for alpha keys then?
 * * Just a thought, could a variable have an array part AND a hash table, where
 *   the array is always as large as the highest index? It could speed things
 *   up ... maybe, even if ths hash table is pretty much O(1) there's lots of
 *   processing and checking going on. Maybe later.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include <ctype.h>
#include <time.h>
#include <assert.h>
#include "prolan.h"
#include "memory_manager.h"
#include "symbol_table.h"

/* Token buffer size. On overflow it'll grow by this value. */
#define TOKEN_BUF_SIZE 10000

/* Function buffer size. On overflow it'll grow by this value. */
#define FUNCTION_BUF_SIZE 64

/* Don't ask, could use the stack. */
#define REGISTER_COUNT 4

/* Only used for arithmetics, needn't be big. */
#define STACK_SIZE 16

/* Some limitations during tokenization. */
#define INLINE_STRING_MAX_LEN 256
#define INLINE_NUMBER_MAX_LEN 32
#define NAME_MAX_LEN 32

/* These are tokenized as single characters. */
#define CHR_EQD         1000 /* := */
#define CHR_INC         1001 /* ++ */
#define CHR_DEC         1002 /* -- */
#define CHR_INCV        1003 /* += */
#define CHR_DECV        1004 /* -= */
#define CHR_MULV        1005 /* *= */
#define CHR_DIVV        1006 /* /= */
#define CHR_LESSOREQ    1009 /* <= */
#define CHR_GREATEROREQ 1010 /* >= */
#define CHR_NOTEQ       1011 /* <> */

/* For things that shouldn't exist, report line in source code. */
#define INTERNAL_ERROR sprintf(sErrorMsg, "Internal error %d", __LINE__); Error()

/*
 * Enum: Reserved
 * --------------
 * Reserved words.
 */
typedef enum {
    RES_IF = 0,
    RES_ELSE,
    RES_WHILE,
    RES_DO,
    RES_FOR,
    RES_BREAK,
    RES_RETURN,
    RES_UNALIAS,
	RES_FREE,
    RES_EXTERN,
    RES_WLN,
    RES_WRITE,
    RES_SLEEP,

    RES_AND,
    RES_OR,
    RES_NOT,

    RES_TOSTRING,
    RES_TONUMBER,
    RES_TIME,
	RES_DEFINED,
    RES_SIZEOF,
    RES_ALIAS,
    RES_FUNCTION,

    /* Statements. */
    RES_DEBUG
} Reserved;

/*
 * Struct: ReservedData
 * --------------------
 * Information about a reserved word.
 */
typedef struct {
    Reserved value;
    const char *name;
} ReservedData;

/*
 * Function: NewReservedData
 * -------------------------
 * Create new reserved data with value.
 *
 * value: value
 *
 * Returns: reserved data
 */
ReservedData *NewReservedData(Reserved value, const char *name);

/*
 * Function: FreeReservedData
 * --------------------------
 * Delete reserved data, for symbol table.
 *
 * data: reserved data
 */
void FreeReservedData(void *data);

/* Symbol table for reserved words. */
static STable *sReservedWords = 0;

/*
 * Enum: TokenType.
 * -----------------
 */
typedef enum {
    TOK_CHARACTER,
    TOK_NUMBER,
    TOK_STRING,
    TOK_NAME,
    TOK_RESERVED,
    TOK_EOF
} TokenType;

/*
 * Struct: Token
 * -------------
 * A token in the input file.
 */
typedef struct {
    /* Source code reference. */
    int line;
    int col; 
    /* Type and value. */
    TokenType type;
    union {
        int i;    /* CHARACTER/RESERVED */
        double d; /* NUMBER */
        char *s;  /* STRING/NAME */
    } value;
    /* For runtime optimizations. */
    int extra;
} Token;

/* Tokens and count. */
static Token *sTokens = 0;
static int sTokenCount = 0;
static int sTokensSize = 0;

/*
 * Struct: Function
 * ----------------
 */
typedef struct {
    Token *token;
    char *params;
    char **paramNames;
    int paramCount;
} Function;

/* Used only when building functions, done once per unique function. */
typedef struct sFunctionParam FunctionParam;
struct sFunctionParam {
    char type;
    char *name;
    FunctionParam *next;
};


/* Functions. */
Function *sFunctions;
int sFunctionSize = 0;
int sFunctionCount = 0;

/*
 * Function: NewFunction
 * ---------------------
 * Return index of new function.
 *
 * Returns: index of new function
 */
int NewFunction();

/*
 * Function: BuildFunction
 * -----------------------
 */
int BuildFunction();

/*
 * Function: ClearFunctions
 * ------------------------
 * Relese all allocated memory for functions.
 */
void ClearFunctions();

/* State. */
static int sBlock = 0;
static int sScope = 0;
static int sLoopScope = 0;
static Token *sToken = 0;
static TokenType sNextType = 0;
static Reserved sNextReserved = 0;
static int sNextChar = 0;
static double sNextNumber = 0.0;
static const char *sNextString = 0;
static char sNextName[1024] = "";
static char sNextKey[1024] = "";
static int sLastLine = 0;
static int sLastCol = 0;

/* Registers. */
static Variable sRegister[REGISTER_COUNT];

/* Stack. */
static Variable sStack[STACK_SIZE];
static int sStackIndex = 0;

/* Error message, displayed by Error. */
static char sErrorMsg[1024] = "";

/*
 * Function: Tokenize
 * ------------------
 * Create tokens from file.
 *
 * Returns: 1 on success, 0 on failure.
 */
int Tokenize(const char *filename);

/*
 * Function: ClearTokens
 * ---------------------
 * Clear tokens.
 */
void ClearTokens();

/*
 * Function: AddToken
 * ------------------
 * Add token.
 *
 * token: token
 */
void AddToken(Token *token);

/*
 * Function: GetNext
 * -----------------
 * Get next token.
 */
void GetNext(int global);

/*
 * Function: ValidUserKey
 * ----------------------
 * Validate sub key supplied by user, var[<string expr>].
 *
 * Returns 1 if valid, else 0.
 */
int ValidateUserKey(const char *key);

/*
 * Function: ExecuteBlock
 * ----------------------
 */
int ExecuteBlock();

/*
 * Function: SkipBlock
 * -------------------
 */
void SkipBlock();

/*
 * Function: Search
 * ----------------
 */
void Search(int c);

/*
 * Function: If
 * ------------
 * if (<bool expr>) {[statements]}
 * [else if (<bool expr>) {[statements]}]
 * ..
 * [else {[statements]}]
 */
int If();

/*
 * Function: While
 * ---------------
 * while (<bool expr>) {[statements]}
 */
int While();

/*
 * Function: Do
 * ------------
 * do {[statements]} while (<bool expr>);
 */
int Do();

/*
 * Function: For
 * -------------
 * for (<variable> = <expr>, <expr>[, <expr>]) {[statements]}
 * for (<variable> : <variable>) {[statements]}
 */
int For();

/*
 * Function: Break
 * ---------------
 */
void Break();

/*
 * Function: Error
 * ---------------
 * Print error message and exit.
 */
void Error();

/*
 * Function: Assignment
 * --------------------
 * Assign expression to variable.
 *
 * dst: variable
 */
void Assignment(Variable *dst);

/*
 * Function: DeepAssignment
 * ------------------------
 * Deep copy variable.
 *
 * dst: variable
 */
void DeepAssignment(Variable *dst);

/*
 * Function: CallFunc
 * ------------------
 */
void CallFunc(Variable *var);

/*
 * Function: Return
 * ----------------
 */
void Return();

/*
 * Function: IncVariable
 * ---------------------
 * Increase variable with expression (+=).
 *
 * var: variable
 */
void IncVariable(Variable *var);

/*
 * Function: DecVariable
 * ---------------------
 * Decrease variable with expression (var -=).
 *
 * var: variable
 */
void DecVariable(Variable *var);

/*
 * Function: MulVariable
 * ---------------------
 * Multiply variable with expression (var *=).
 *
 * var: variable
 */
void MulVariable(Variable *var);

/*
 * Function: DivVariable
 * ---------------------
 * Divide variable with expression (var /=).
 *
 * var: variable
 */
void DivVariable(Variable *var);

/*
 * Function: VariableInc
 * ---------------------
 * Increase variable by 1 (++var).
 *
 * var: variable
 */
void VariableInc(Variable *var);

/*
 * Function: VariableDec
 * ---------------------
 * Decrease variable by 1 (--var).
 *
 * var: variable
 */
void VariableDec(Variable *var);

/*
 * Function: CheckPostMod
 * ----------------------
 * Apply any post modifiers (++ and --) on variable.
 *
 * var: variable
 */
void CheckPostMod(Variable *var);

/*
 * Function: Expect
 * ----------------
 * Error if character is missing.
 */
void Expect(int c);

/*
 * Function: EndStatement
 * ----------------------
 * Error if ; is missing.
 */
void EndStatement();

/*
 * Function: Expression
 * --------------------
 * Evaluate expression and put result in register 0.
 */ 
void Expression();

/*
 * Function: Term
 * --------------
 */
void Term();

/*
 * Function: SignedFactor
 * ----------------------
 */
void SignedFactor();

/*
 * Function: Factor
 * ----------------
 */
void Factor();

/*
 * Function: BoolExpression
 * ------------------------
 */
void BoolExpression();

/*
 * Function: SingleBool
 * --------------------
 */
void SingleBool();

/*
 * Function: BoolTerm
 * ------------------
 */
void BoolTerm();

/*
 * Function: IsRelOp
 * -----------------
 */
int IsRelOp(int c);

/*
 * Function: Push
 * --------------
 * Push variable to stack.
 *
 * var: variable
 */
void Push(Variable *var);

/*
 * Function: Pop
 * -------------
 * Pop variable from stack.
 */
void Pop(Variable *var);

/*
 * Function: Add
 * -------------
 */
void Add();

/*
 * Function: Sub
 * -------------
 */
void Sub();

/*
 * Function: Mul
 * -------------
 */
void Mul();

/*
 * Function: Div
 * -------------
 */
void Div();

/*
 * Function: AddString
 * -------------------
 */
void AddString();

/*
 * Function: Wln
 * -------------
 * wln(<string expr>);
 */
void Wln();

/*
 * Function: Write
 * ---------------
 * write(<string expr>);
 */
void Write();

/*
 * Function ToString
 * -----------------
 * tostring(<expr>)
 */
void ToString();

/*
 * Function: ToNumber
 * ------------------
 * tonumber(<expr>)
 */
void ToNumber();

/*
 * Function: SizeOf
 * ----------------
 * sizeof(<variable>)
 */
void SizeOf();

/*
 * Function: Time
 * --------------
 * time(), time in milliseconds since system boot.
 */
void Time();

/*
 * Function: Sleep
 * ---------------
 * sleep(milliseconds);
 */
void Sleep();

/*
 * Function: Free
 * --------------
 * delete <variable>;
 */
void Free();

/*
 * Function: Extern
 * ----------------
 * extern <var>[, <var> ...];
 */
void Extern();

/*
 * Function: Unalias
 * -----------------
 * unref <var>;
 */
void Unalias();

/*
 * Function: PL_Init
 * -----------------
 * Init prolan.
 */
void PL_Init() {
    if (!sReservedWords) {
        sReservedWords = ST_Init(FreeReservedData, 0);
        ST_AddData(sReservedWords, "if", NewReservedData(RES_IF, "if"));
        ST_AddData(sReservedWords, "else", NewReservedData(RES_ELSE, "else"));
        ST_AddData(sReservedWords, "while", NewReservedData(RES_WHILE, "while"));
        ST_AddData(sReservedWords, "do", NewReservedData(RES_DO, "do"));
        ST_AddData(sReservedWords, "for", NewReservedData(RES_FOR, "for"));
        ST_AddData(sReservedWords, "break", NewReservedData(RES_BREAK, "break"));
        ST_AddData(sReservedWords, "and", NewReservedData(RES_AND, "and"));
        ST_AddData(sReservedWords, "or", NewReservedData(RES_OR, "or"));
        ST_AddData(sReservedWords, "not", NewReservedData(RES_NOT, "not"));
        ST_AddData(sReservedWords, "write", NewReservedData(RES_WRITE, "write"));
        ST_AddData(sReservedWords, "wln", NewReservedData(RES_WLN, "wln"));
        ST_AddData(sReservedWords, "tostring", NewReservedData(RES_TOSTRING, "tostring"));
        ST_AddData(sReservedWords, "tonumber", NewReservedData(RES_TONUMBER, "tonumber"));
        ST_AddData(sReservedWords, "time", NewReservedData(RES_TIME, "time"));
        ST_AddData(sReservedWords, "sleep", NewReservedData(RES_SLEEP, "sleep"));
        ST_AddData(sReservedWords, "defined", NewReservedData(RES_DEFINED, "defined"));
        ST_AddData(sReservedWords, "sizeof", NewReservedData(RES_SIZEOF, "sizeof"));
        ST_AddData(sReservedWords, "delete", NewReservedData(RES_FREE, "delete"));
        ST_AddData(sReservedWords, "ref", NewReservedData(RES_ALIAS, "ref"));
        ST_AddData(sReservedWords, "unref", NewReservedData(RES_UNALIAS, "unref"));
        ST_AddData(sReservedWords, "function", NewReservedData(RES_FUNCTION, "function"));
        ST_AddData(sReservedWords, "return", NewReservedData(RES_RETURN, "return"));
        ST_AddData(sReservedWords, "extern", NewReservedData(RES_EXTERN, "extern"));
        ST_AddData(sReservedWords, "debug", NewReservedData(RES_DEBUG, "debug"));
    }
    MM_Init();
}

/*
 * Function: PL_Terminate
 * ----------------------
 * Terminate prolan.
 */
void PL_Terminate() {
    if (sReservedWords) {
        ST_Clear(sReservedWords);
        free(sReservedWords);
        sReservedWords = 0;
    }
    MM_Terminate();
}

/*
 * Function: PL_Run
 * ----------------
 * Execute a prolan source code file.
 *
 * filename: source code filename
 *
 * Returns: 1 on success, 0 on failure.
 */
int PL_Run(const char *filename) {
    if (!Tokenize(filename)) {
        ClearTokens();
        return 0;
    }
    sToken = sTokens;
    GetNext(0);
    ExecuteBlock();
    ClearTokens();
    ClearFunctions();
    return 1;
}

/*
 * Function: NewReservedData
 * -------------------------
 * Create new reserved data with value.
 *
 * value: value
 *
 * Returns: reserved data
 */
ReservedData *NewReservedData(Reserved value, const char *name) {
    ReservedData *rd = (ReservedData *)malloc(sizeof(ReservedData));
    rd->value = value;
    rd->name = name;
    return rd;
}

/*
 * Function: FreeReservedData
 * --------------------------
 * Delete reserved data, for symbol table.
 *
 * data: reserved data
 */
void FreeReservedData(void *data) {
    ReservedData *rd = (ReservedData *)data;
    free(rd);
}

/*
 * Function: Tokenize
 * ------------------
 * Create tokens from file. Does some basic error checking.
 *
 * Returns: 1 on success, 0 on failure.
 */
int Tokenize(const char *filename) {
    FILE *file = fopen(filename, "r");
    int line = 1;
    int col = 1;
    int c;
    Token token;
    
    /* Used for runtime optimizations, such as skips. */
    token.extra = -1;

    if (!file) return 0;

    c = fgetc(file);
    while (c != EOF) {
        /* Skip whitespace. */
        while (c == ' ' || c == '\t' || c == '\n' || c == '\r') {
            if (c == '\n') {
                line++;
                col = 0;
            }
            c = fgetc(file);
            col++;
        }
        if (c == EOF) break;

        /* Number. */
        if (isdigit(c)) {
            char number[INLINE_NUMBER_MAX_LEN];
            int len = 0;
            int point = 0;
            int startCol = col;
            do {
                number[len++] = c;
                if (len == INLINE_NUMBER_MAX_LEN) {
                    printf("ERROR (%d:%d): Numeric constant too long\n", line, startCol);
                    fclose(file);
                    return 0;
                }
                c = fgetc(file);
                col++;
            } while (isdigit(c) || (c == '.' && point++ == 0));
            number[len] = '\0';
            token.type = TOK_NUMBER;
            token.value.d = atof(number);
            token.line = line;
            token.col = startCol;
            AddToken(&token);
        }
        /* Variable or reserved. */
        else if (isalpha(c) || c == '_') {
            char str[NAME_MAX_LEN];
            int len = 0;
            int startCol = col;
            ReservedData *rd;
            do {
                str[len++] = c;
                if (len == NAME_MAX_LEN) {
                    printf("ERROR (%d:%d): Name too long\n", line, startCol);
                    fclose(file);
                    return 0;
                }
                /* Add check on 'len'. */
                c = fgetc(file);
                col++;
            } while (isalpha(c) || isdigit(c) || c == '_');
            str[len] = '\0';
            rd = (ReservedData *)ST_GetData(sReservedWords, str);
            token.line = line;
            token.col = startCol;
            if (rd) {
                token.type = TOK_RESERVED;
                token.value.i = rd->value;
            }
            else {
                token.type = TOK_NAME;
                token.value.s = strdup(str);
            }
            AddToken(&token);
        }
        /* String. */
        else if (c == '\"') {
            char str[INLINE_STRING_MAX_LEN];
            int len = 0;
            int startCol = col;
            /* You should handle escape codes in this loop. */
            c = fgetc(file);
            col++;
            while (c != '\"') {
                str[len++] = c;
                if (len == INLINE_STRING_MAX_LEN) {
                    printf("ERROR (%d:%d): String constant too long\n", line, startCol);
                    fclose(file);
                    return 0;
                }
                c = fgetc(file);
                col++;
                if (c == EOF) {
                    printf("ERROR (%d:%d): Unterminated string\n", line, startCol);
                    fclose(file);
                    return 0;
                }
            }
            str[len] = '\0';
            token.type = TOK_STRING;
            token.value.s = strdup(str);
            token.line = line;
            token.col = startCol;
            AddToken(&token);
            c = fgetc(file);
            col++;
        }
        /* Character. */
        else {
            int comment = 0;
            token.type = TOK_CHARACTER;
            token.value.i = c;
            token.line = line;
            token.col = col;
            c = fgetc(file);
            col++;
            if (token.value.i == '+') {
                if (c == '+') {
                    token.value.i = CHR_INC; 
                    c = fgetc(file);
                    col++;
                }
                else if (c == '=') {
                    token.value.i = CHR_INCV;
                    c = fgetc(file);
                    col++;
                }
            }
            else if (token.value.i == '-') {
                if (c == '-') {
                    token.value.i = CHR_DEC;
                    c = fgetc(file);
                    col++;
                }
                else if (c == '=') {
                    token.value.i = CHR_DECV;
                    c = fgetc(file);
                    col++;
                }
            }
            else if (token.value.i == '*') {
                if (c == '=') {
                    token.value.i = CHR_MULV;
                    c = fgetc(file);
                    col++;
                }
            }
            else if (token.value.i == '/') {
                if (c == '=') {
                    token.value.i = CHR_DIVV;
                    c = fgetc(file);
                    col++;
                }
                else if (c == '/') {
                    comment = 1;
                    c = fgetc(file);
                    while (!(c == EOF || c == '\n')) {
                        c = fgetc(file);
                    }
                }
                else if (c == '*') {
                    int prevC = 0;
                    int startCol = col;
                    comment = 1;
                    c = fgetc(file);
                    col++;
                    do {
                        if (prevC == '*' && c == '/') {
                            c = fgetc(file);
                            col++;
                            break;
                        }
                        else if (c == '\n') {
                            line++;
                            col = 0;
                        }
                        else if (c == EOF) {
                            printf("ERROR (%d:%d): Unterminated comment\n", line, startCol);
                            fclose(file);
                            return 0;
                        }
                        prevC = c;
                        c = fgetc(file);
                        col++;
                    } while (1);
                }
            }
            else if (token.value.i == '<') {
                if (c == '=') {
                    token.value.i = CHR_LESSOREQ;
                    c = fgetc(file);
                    col++;
                }
                else if (c == '>') {
                    token.value.i = CHR_NOTEQ;
                    c = fgetc(file);
                    col++;
                }
            }
            else if (token.value.i == '>') {
                if (c == '=') {
                    token.value.i = CHR_GREATEROREQ;
                    c = fgetc(file);
                    col++;
                }
            }
            else if (token.value.i == ':') {
                if (c == '=') {
                    token.value.i = CHR_EQD;
                    c = fgetc(file);
                    col++;
                }
            }
            if (!comment) {
                AddToken(&token);
            }
        }
    }
    token.type = TOK_EOF;
    token.line = line;
    token.col = col;
    AddToken(&token);

    fclose(file);
    
    return 1;
}

/*
 * Function: AddToken
 * ------------------
 * Add token.
 *
 * token: token
 */
void AddToken(Token *token) {
    if (sTokens == 0) {
        sTokensSize = TOKEN_BUF_SIZE;
        sTokens = (Token *)malloc(sizeof(Token)*sTokensSize);
        sTokenCount = 0;
    }
    else if (sTokenCount == sTokensSize) {
        Token *newTokens;
        newTokens = (Token *)malloc(sizeof(Token)*(sTokensSize + TOKEN_BUF_SIZE));
        memcpy(newTokens, sTokens, sizeof(Token)*sTokensSize);
        sTokensSize += TOKEN_BUF_SIZE;
        free(sTokens);
        sTokens = newTokens;
    }
    sTokens[sTokenCount] = *token;
    sTokens[sTokenCount++].extra = 0;
}

/*
 * Function: ClearTokens
 * ---------------------
 * Clear tokens.
 */
void ClearTokens() {
    if (sTokens) {
        int i;
        for (i = 0; i < sTokenCount; i++) {
            if (sTokens[i].type == TOK_STRING || sTokens[i].type == TOK_NAME) {
                free(sTokens[i].value.s);
            }
        }       
        free(sTokens);
        sTokens = 0;
        sTokensSize = 0;
        sTokenCount = 0;
    }
}

/*
 * Function: NewFunction
 * ---------------------
 * Return index of new function.
 *
 * Returns: index of new function
 */
int NewFunction() {
    if (sFunctions == 0) {
        sFunctionSize = FUNCTION_BUF_SIZE;
        sFunctions = (Function *)malloc(sizeof(Function)*sFunctionSize);
        sFunctionCount = 0;
    }
    else if (sFunctionCount == sFunctionSize) {
        Function *newFunctions;
        newFunctions = (Function *)malloc(sizeof(Function)*(sFunctionSize + FUNCTION_BUF_SIZE));
        memcpy(newFunctions, sFunctions, sizeof(Function)*sFunctionSize);
        sFunctionSize += FUNCTION_BUF_SIZE;
        free(sFunctions);
        sFunctions = newFunctions;
    }
    sFunctions[sFunctionCount].params = 0;
    sFunctions[sFunctionCount].paramNames = 0;
    sFunctions[sFunctionCount].paramCount = 0;
    return sFunctionCount++;
}

/*
 * Function: ParamListContains
 * ---------------------------
 */
int ParamListContains(FunctionParam *list, const char *name) {
    while (list) {
        if (strcmp(list->name, name) == 0) return 1;
        list = list->next;
    }
    return 0;
}

/*
 * Function: FreeParamList
 * -----------------------
 */
void FreeParamList(FunctionParam *list, int freeName) {
    while (list) {
        FunctionParam *next = list->next;
        if (freeName) {
            free(list->name);
        }
        free(list);
        list = next;
    }
}

/*
 * Function: BuildFunction
 * -----------------------
 */
int BuildFunction() {
    int function = NewFunction();
    FunctionParam *paramList = 0;
    FunctionParam *param = 0;
    int paramCount = 0;

    /* Store function index + 1 in token, so that a function is never built
       twice. */
    sToken->extra = function + 1;
    GetNext(0); /* function */
    
    /*Expect('(');*/
    /* We need to call GetNext with global set to true so that parameter names
       are not affected by the current scope. Therefore we can't use Expect,
       since it doesn't set the global flag. */
    if (!(sNextType == TOK_CHARACTER && sNextChar == '(')) {
        sprintf(sErrorMsg, "Expected (");
        Error();
    }
    GetNext(1);
    while (!(sNextType == TOK_CHARACTER && sNextChar == ')')) {
        char type;
        if (sNextType == TOK_RESERVED && sNextReserved == RES_ALIAS) {
            GetNext(1);
            type = 'r';
        }
        else {
            type = 'v';
        }
        if (sNextType != TOK_NAME) {
            FreeParamList(paramList, 1);
            sprintf(sErrorMsg, "Expected parameter name");
            Error();
        }
        if (ParamListContains(paramList, sNextKey)) {
            FreeParamList(paramList, 1);
            sprintf(sErrorMsg, "Duplicate parameter name \"%s\"", sNextKey);
            Error();
        }
        if (paramList == 0) {
            paramList = (FunctionParam *)malloc(sizeof(FunctionParam));
            param = paramList;
        }
        else {
            param->next = (FunctionParam *)malloc(sizeof(FunctionParam));
            param = param->next;
        }
        param->type = type;
        param->name = strdup(sNextKey); /* Wait, what if key is foo.bar.shit[1]? */
        param->next = 0;
        paramCount++;        
        GetNext(1);

        if (sNextType == TOK_CHARACTER) {
            if (sNextChar == ',') {
                GetNext(1);
            }
            else if (sNextChar != ')') {
                FreeParamList(paramList, 1);
                sprintf(sErrorMsg, "Expected , or )");
                Error();
            }
        }
        else {
            FreeParamList(paramList, 1);
            sprintf(sErrorMsg, "Expected , or )");
            Error();
        }
    }
    GetNext(0); /* ) */
    sFunctions[function].token = sToken;
    sFunctions[function].paramCount = paramCount;
    if (paramCount) {
        int index = 0;
        sFunctions[function].params = (char *)malloc(sizeof(char)*(paramCount + 1));
        sFunctions[function].paramNames = (char **)malloc(sizeof(char *)*paramCount);
        param = paramList;
        while (param) {
            sFunctions[function].params[index] = param->type;
            sFunctions[function].paramNames[index] = param->name;
            param = param->next;
            index++;
        }
        sFunctions[function].params[index] = '\0';
        FreeParamList(paramList, 0);
    }
    else {
        sFunctions[function].params = 0;
        sFunctions[function].paramNames = 0;
    }

    /*if (paramCount) {
        int i;
        printf("function %d params:\n", function);
        for (i = 0; i < paramCount; i++) {
            printf(" (%c)%s\n", sFunctions[function].params[i], sFunctions[function].paramNames[i]);
        }
    }*/

    return function;
}

int BuildFunctionOld() {
    int function = NewFunction();
    FunctionParam *paramList = 0;
    FunctionParam *param = 0;
    int paramCount = 0;

    /* Store function index + 1 in token, so that a function is never built
       twice. */
    sToken->extra = function + 1;
    GetNext(0); /* function */
    
    Expect('(');
    while (!(sNextType == TOK_CHARACTER && sNextChar == ')')) {
        char type;
        if (sNextType == TOK_RESERVED && sNextReserved == RES_ALIAS) {
            GetNext(0);
            type = 'r';
        }
        else {
            type = 'v';
        }
        if (sNextType != TOK_NAME) {
            FreeParamList(paramList, 1);
            sprintf(sErrorMsg, "Expected parameter name");
            Error();
        }
        if (ParamListContains(paramList, sNextKey)) {
            FreeParamList(paramList, 1);
            sprintf(sErrorMsg, "Duplicate parameter name \"%s\"", sNextKey);
            Error();
        }
        if (paramList == 0) {
            paramList = (FunctionParam *)malloc(sizeof(FunctionParam));
            param = paramList;
        }
        else {
            param->next = (FunctionParam *)malloc(sizeof(FunctionParam));
            param = param->next;
        }
        param->type = type;
        param->name = strdup(sNextKey); /* Wait, what if key is foo.bar.shit[1]? */
        param->next = 0;
        paramCount++;        
        GetNext(0);

        if (sNextType == TOK_CHARACTER) {
            if (sNextChar == ',') {
                GetNext(0);
            }
            else if (sNextChar != ')') {
                FreeParamList(paramList, 1);
                sprintf(sErrorMsg, "Expected , or )");
                Error();
            }
        }
        else {
            FreeParamList(paramList, 1);
            sprintf(sErrorMsg, "Expected , or )");
            Error();
        }
    }
    GetNext(0); /* ) */
    sFunctions[function].token = sToken;
    sFunctions[function].paramCount = paramCount;
    if (paramCount) {
        int index = 0;
        sFunctions[function].params = (char *)malloc(sizeof(char)*(paramCount + 1));
        sFunctions[function].paramNames = (char **)malloc(sizeof(char *)*paramCount);
        param = paramList;
        while (param) {
            sFunctions[function].params[index] = param->type;
            sFunctions[function].paramNames[index] = param->name;
            param = param->next;
            index++;
        }
        sFunctions[function].params[index] = '\0';
        FreeParamList(paramList, 0);
    }
    else {
        sFunctions[function].params = 0;
        sFunctions[function].paramNames = 0;
    }

    /*if (paramCount) {
        int i;
        printf("function %d params:\n", function);
        for (i = 0; i < paramCount; i++) {
            printf(" (%c)%s\n", sFunctions[function].params[i], sFunctions[function].paramNames[i]);
        }
    }*/

    return function;
}

/*
 * Function: ClearFunctions
 * ------------------------
 * Relese all allocated memory for functions.
 */
void ClearFunctions() {
    if (sFunctions) {
        int i;
        for (i = 0; i < sFunctionCount; i++) {
            if (sFunctions[i].paramCount) {
                int j;
                free(sFunctions[i].params);
                for (j = 0; j < sFunctions[i].paramCount; j++) {
                    free(sFunctions[i].paramNames[j]);
                }
                free(sFunctions[i].paramNames);
            }
        }       
        free(sFunctions);
        sFunctions = 0;
        sFunctionSize = 0;
        sFunctionCount = 0;
    }
}

/*
 * Function: ExecuteBlock
 * ----------------------
 */
int ExecuteBlock() {
    sBlock++;
    do {
        if (sNextType == TOK_EOF) {
            if (sBlock > 1) {
                sprintf(sErrorMsg, "Unexpected end of file");
                Error();
            }
            else {
                break;
            }
        }
        else if (sNextType == TOK_RESERVED) {
            int res;
            switch (sNextReserved) {
            case RES_IF:
                res = If();
                if (res) {
                    sBlock--;
                    if (res == 1) SkipBlock();
                    return res;
                }
                break;
            case RES_WHILE:
                res = While();
                if (res) {
                    sBlock--;
                    if (res == 1) SkipBlock();
                    return res;
                }
                break;
            case RES_DO:
                res = Do();
                if (res) {
                    sBlock--;
                    if (res == 1) SkipBlock();
                    return res;
                }
                break;
            case RES_FOR:
                res = For();
                if (res) {
                    sBlock--;
                    if (res == 1) SkipBlock();
                    return res;
                }
                break;
            case RES_BREAK:
                Break();
                sBlock--;
                return 1;
                break;
            case RES_RETURN:
                Return();
                sBlock--;
                return 2;
                break;
            case RES_WLN:
                Wln();
                break;
            case RES_WRITE:
                Write();
                break;
            case RES_EXTERN:
                Extern();
                break;
            case RES_FREE:
                Free();
                break;
            case RES_UNALIAS:
                Unalias();
                break;
            case RES_SLEEP:
                Sleep();
                break;
            default:
                sprintf(sErrorMsg, "Syntax error");
                Error();
                break;
            }
        }
        else if (sNextType == TOK_NAME) {
            /* Assignment, inc/dec or function call. */
            Variable *var = MM_CreateVar(sNextKey);
            GetNext(0);
            if (sNextType == TOK_CHARACTER) {
                switch (sNextChar) {
                case '=':
                    Assignment(var);
                    break;
                case CHR_EQD:
                    DeepAssignment(var);
                    break;
                case CHR_INC:
                    VariableInc(var);
                    GetNext(0);
                    break;
                case CHR_DEC:
                    VariableDec(var);
                    GetNext(0);
                    break;
                case CHR_INCV:
                    IncVariable(var);
                    break;
                case CHR_DECV:
                    DecVariable(var);
                    break;
                case CHR_MULV:
                    MulVariable(var);
                    break;
                case CHR_DIVV:
                    DivVariable(var);
                    break;
                case '(':
                    CallFunc(var);
                    MM_DeleteVar("_");
                    break;
                default:
                    sprintf(sErrorMsg, "Syntax error");
                    Error();
                }
                EndStatement();
            }
            else if (sNextType == TOK_RESERVED && sNextReserved == RES_ALIAS) {
                GetNext(0);
                if (sNextType == TOK_NAME) {
                    Var_SetAlias(var, sNextKey);
                    if (var->type != VAR_ALIAS) {
                        sprintf(sErrorMsg, "Can't create alias for undefined");
                        Error();
                    }
                }
                else {
                    sprintf(sErrorMsg, "Expected variable");
                    Error();
                }
                GetNext(0);
                EndStatement();
            }
            else {
                sprintf(sErrorMsg, "Syntax error");
                Error();
            }
        }
        else if (sNextType == TOK_CHARACTER) {
            if (sNextChar == CHR_INC) {
                GetNext(0);
                if (sNextType == TOK_NAME) {
                    Variable *var = MM_CreateVar(sNextKey);
                    VariableInc(var);
                    GetNext(0);
                    EndStatement();
                }
                else {
                    sprintf(sErrorMsg, "Expected variable");
                    Error();
                }
            }
            else if (sNextChar == CHR_DEC) {
                GetNext(0);
                if (sNextType == TOK_NAME) {
                    Variable *var = MM_CreateVar(sNextKey);
                    VariableDec(var);
                    GetNext(0);
                    EndStatement();
                }
                else {
                    sprintf(sErrorMsg, "Expected variable");
                    Error();
                }
            }
            else if (sNextChar == '}' && sBlock > 1) {
                break;
            }
            else {
                sprintf(sErrorMsg, "Syntax error");
                Error();
            }
        }
        else {
            sprintf(sErrorMsg, "Syntax error");
            Error();
        }
    } while (1);
    sBlock--;
    return 0;
}

/*
 * Function: SkipBlock
 * -------------------
 */
void SkipBlock() {
    int blockc = 1;
    if (sNextType == TOK_CHARACTER && sNextChar == '}') {
        GetNext(0);
        return;
    }
 
    do {
        if (sToken->type == TOK_CHARACTER && sToken->value.i == '}') {
            if (--blockc == 0) {
                sToken++;
                break;
            }
        }
        else if (sToken->type == TOK_CHARACTER && sToken->value.i == '{') {
            blockc++;
        }
        else if (sToken->type == TOK_EOF) {
            sLastLine = sToken->line;
            sLastCol = sToken->col;
            sprintf(sErrorMsg, "Expected } here");
            Error();
        }
        sToken++;    
    } while (1);
    GetNext(0);
}

/*
 * Function: Search
 * ----------------
 */
void Search(int c) {
    if (sNextType == TOK_CHARACTER && sNextChar == c) {
        return;
    }

    do {
        if (sToken->type == TOK_CHARACTER && sToken->value.i == c) {
            sLastLine = sToken->line;
            sLastCol = sToken->col;
            sNextType = TOK_CHARACTER;
            sNextChar = c;
            sToken++;
            break;
        }
        else if (sToken->type == TOK_EOF) {
            sLastLine = sToken->line;
            sLastCol = sToken->col;
            sprintf(sErrorMsg, "Expected %c", c);
            Error();
        }
        sToken++;
    } while (1);
}

/*
 * Function: If
 * ------------
 * if (<bool expr>) {[statements]}
 * [else if (<bool expr>) {[statements]}]
 * ..
 * [else {[statements]}]
 */
int If() {
    int catched = 0;
    GetNext(0);
    Expect('(');
    BoolExpression();
    Expect(')');
    if (sRegister[0].value.i) {
        int res;
        catched = 1;
        Expect('{');
        res = ExecuteBlock();
        if (res) return res;
        Expect('}');
    }
    else {
        /* The full skip process should only be done once, an index could be
           stored in the { token that points to the skip destination. */
        Expect('{');
        SkipBlock();
    }
    /* Loop while next is 'else' or 'else if'. */
    while (sNextType == TOK_RESERVED && sNextReserved == RES_ELSE) {
        GetNext(0);
        /* 'else if' */
        if (sNextType == TOK_RESERVED && sNextReserved == RES_IF) {
            GetNext(0);
            if (catched) {
                Search('{');
                Expect('{');
                SkipBlock();
            }
            else {
                Expect('(');
                BoolExpression();
                Expect(')');
                if (sRegister[0].value.i) {
                    int res;
                    catched = 1;
                    Expect('{');
                    res = ExecuteBlock();
                    if (res) return res;
                    Expect('}');
                }
                else {
                    Expect('{');
                    SkipBlock();
                }
            }
        }
        /* 'else' */
        else {
            Expect('{');
            if (catched) {
                SkipBlock();
            }
            else {
                int res = ExecuteBlock();
                if (res) return res;
                Expect('}');
            }
            break;
        }
    }
    return 0;
}

/*
 * Function: While
 * ---------------
 * while (<bool expr>) {[statements]}
 */
int While() {
    Token *loop = sToken;
    sLoopScope++;
    do {
        int res;
        GetNext(0);
        Expect('(');
        BoolExpression();
        Expect(')');
        if (sRegister[0].value.i == 0) break;
        Expect('{');
        res = ExecuteBlock();
        if (res == 1) {
            SkipBlock();
            sLoopScope--;
            return 0;
        }
        else if (res == 2) {
            sLoopScope--;
            return 2;
        }
        Expect('}');
        sToken = loop;
    } while (1);
    Expect('{');
    SkipBlock();
    sLoopScope--;
    return 0;
}

/*
 * Function: Do
 * ------------
 * do {[statements]} while (<bool expr>);
 */
int Do() {
    Token *loop = sToken;
    sLoopScope++;
    do {
        int res;
        GetNext(0);
        Expect('{');
        res = ExecuteBlock();
        if (res == 1) {
            SkipBlock();
            Search(';');
            EndStatement();
            break;
        }
        else if (res == 2) {
            sLoopScope--;
            return 2;
        }
        Expect('}');
        if (!(sNextType == TOK_RESERVED && sNextReserved == RES_WHILE)) {
            sprintf(sErrorMsg, "Expected while");
            Error();
        }
        GetNext(0);
        Expect('(');
        BoolExpression();
        Expect(')');
        /* Yeah, end statement here rather than outside of the loop, because
           the ; character can be missing without error on break (that searches
           for the first best ;, which might be after another statement ... */
        EndStatement();
        if (sRegister[0].value.i == 0) break;
        sToken = loop;
    } while (1);
    sLoopScope--;
    return 0;
}

/*
 * Function: For
 * -------------
 * for (<variable> = <expr>, <expr>[, <expr>]) {[statements]}
 * for (<variable> : <variable>) {[statements]}
 */
int For() {
    Variable *loopVar;

    sLoopScope++;
    GetNext(0);
    Expect('(');
    if (sNextType != TOK_NAME) {
        sprintf(sErrorMsg, "Expected variable");
        Error();    
    }
    loopVar = MM_CreateVar(sNextKey);
    GetNext(0);

    /* for (<variable> : <variable>) {[statements]} */
    if (sNextType == TOK_CHARACTER && sNextChar == ':') {
        char key[512];
        int baseLength;
        KeyList *keyList;
        KeyList *currentKey;
        Token *loop;

        GetNext(0);
        if (sNextType != TOK_NAME) {
            sprintf(sErrorMsg, "Expected variable");
            Error();
        }
        strcpy(key, sNextKey);
        strcat(key, ".");
        baseLength = strlen(key);
        keyList = MM_GetKeys(sNextKey);
        GetNext(0);
        loop = sToken;
        Expect(')');
        currentKey = keyList;
        do {
            int res;
            if (!currentKey) {
                Expect('{');
                SkipBlock();
                break;
            }
            key[baseLength] = '\0';
            strcat(key, currentKey->key);
            Var_SetAlias(loopVar, key);

            Expect('{');
            res = ExecuteBlock();
            Var_Free(loopVar);
            if (res == 1) {
                SkipBlock();
                sLoopScope--;
                MM_FreeKeys(keyList); /* DUUDE, MM should not have a static key, inner loops .. */
                return 0;
            }
            else if (res == 2) {
                MM_FreeKeys(keyList);
                return 2;
            }
            Expect('}');
            sToken = loop;
            currentKey = currentKey->next;
            GetNext(0);
            
        } while(1);
        MM_FreeKeys(keyList);
    }
    /* for (<variable> = <expr>, <expr>[, <expr>])
       {[statements]} */
    else if (sNextType == TOK_CHARACTER && sNextChar == '=') {
        Token *loop;
        double dstValue;
        double step;
        GetNext(0);
        Expression();
        if (sRegister[0].type != VAR_NUMBER) {
            sprintf(sErrorMsg, "Expected numeric expression");
            Error();
        }
        Var_SetNumber(loopVar, sRegister[0].value.d);
        Expect(',');
        Expression();
        if (sRegister[0].type != VAR_NUMBER) {
            sprintf(sErrorMsg, "Expected numeric expression");
            Error();
        }
        dstValue = sRegister[0].value.d;

        if (sNextType == TOK_CHARACTER && sNextChar == ',') {
            GetNext(0);
            Expression();
            if (sRegister[0].type != VAR_NUMBER) {
                sprintf(sErrorMsg, "Expected numeric expression");
                Error();
            }
            step = sRegister[0].value.d;
        }
        else {
            step = 1;
        }
        if (dstValue > loopVar->value.d) {
            step = fabs(step);
        }
        else if (dstValue < loopVar->value.d) {
            step = -fabs(step);
        }
        else {
            step = 0;
        }

        loop = sToken;
        Expect(')');
        do {
            int res;
            Expect('{');
            res = ExecuteBlock();
            if (res == 1) {
                SkipBlock();
                sLoopScope--;
                return 0;
            }
            else if (res == 2) {
                sLoopScope--;
                return 2;
            }
            Expect('}');
            loopVar->value.d += step;
            if (step < 0) {
                if (loopVar->value.d < dstValue) break;
            }
            else if (step > 0) {
                if (loopVar->value.d > dstValue) break;
            }
            else {
                break;
            }
            sToken = loop;
            GetNext(0);
        } while (1);                    
    }
    else {
        sprintf(sErrorMsg, "Syntax error");
        Error();
    }

    sLoopScope--;

    return 0;
}

/*
 * Function: Break
 * ---------------
 */
void Break() {
    if (sLoopScope <= 0) {
        sprintf(sErrorMsg, "Unexpected break");
        Error();
    }
    GetNext(0);
    EndStatement();
}

/*
 * Function: GetNext
 * -----------------
 * Get next token.
 */
void GetNext(int global) {
    char key[512] = "";
    char name[512] = "";
    int lastLine = sToken->line;
    int lastCol = sToken->col;

    sNextType = sToken->type;
    switch (sNextType) {
    case TOK_CHARACTER:
        sNextChar = sToken->value.i;
        sToken++;
        break;
    case TOK_NUMBER:
        sNextNumber = sToken->value.d;
        sToken++;
        break;
    case TOK_STRING:
        sNextString = sToken->value.s;
        sToken++;
        break;
    case TOK_RESERVED:
        sNextReserved = sToken->value.i;
        sToken++;
        break;
    case TOK_EOF:
        sNextType = TOK_EOF;
        sToken++;
        break;
    case TOK_NAME:
        /* Change compared to naalaa, globals can only be defined at first level. */
        if (sScope > 0 && !global) {
            sprintf(key, "%d.", sScope);
        }
        while (sToken->type == TOK_NAME || (sToken->type == TOK_CHARACTER &&
                                            (sToken->value.i == '.' ||
                                             sToken->value.i == '['))) {
            if (sToken->type == TOK_NAME) {
                strcat(name, sToken->value.s);
                strcat(key, sToken->value.s);
                sToken++;
                if (sToken->type != TOK_CHARACTER) break;
            }
            else if (sToken->type == TOK_CHARACTER) {
                if (sToken->value.i == '.') {
                    strcat(name, ".");
                    strcat(key, ".");
                    sToken++;
                    if (sToken->type != TOK_NAME) {
                        sprintf(sErrorMsg, "Expected name");
                        Error(0);
                    }
                }
                /* [ */
                else {
                    strcat(name, "[");
                    strcat(key, ".");
                    sToken++;
                    GetNext(0);
                    Expression();
                    if (sRegister[0].type == VAR_NUMBER) {
                        char tmp[10];
                        sprintf(tmp, "%d", (int)sRegister[0].value.d);
                        strcat(name, tmp);
                        strcat(key, tmp);
                    }
                    else if (sRegister[0].type == VAR_STRING) {
                        if (!ValidateUserKey(sRegister[0].value.s)) {
                            sprintf(sErrorMsg, "Invalid key \"%s\"", sRegister[0].value.s);
                            Error();
                        }
                        else {
                            strcat(name, sRegister[0].value.s);
                            strcat(key, sRegister[0].value.s);
                            Var_Free(&sRegister[0]);           
                        }
                    }
                    else if (sRegister[0].type == VAR_UNDEFINED) {
                        sprintf(sErrorMsg, "Can't index array with undefined");
                        Error();
                    }
                    else if (sRegister[0].type == VAR_FUNCTION) {
                        sprintf(sErrorMsg, "Can't index array with function");
                        Error();
                    }
                    else {
                        sprintf(sErrorMsg, "Syntax error");
                        Error();
                    }
                    if (!(sNextType == TOK_CHARACTER && sNextChar == ']')) {
                        sprintf(sErrorMsg, "Expected ]");
                        Error();
                    }
                    strcat(name, "]");
                }
            }
        }
        /* Change from naalaa, scope at top. */
        strcpy(sNextName, name);
        strcpy(sNextKey, key);
        sNextType = TOK_NAME;
        break;
    }
    sLastLine = lastLine;
    sLastCol = lastCol;
}

/*
 * Function: ValidUserKey
 * ----------------------
 * Validate sub key supplied by user, var[<string expr>].
 *
 * Returns 1 if valid, else 0.
 */
int ValidateUserKey(const char *key) {
    const char *c = key;
    int type = 0;
    while (*c != '\0') {
        if (type == 0) {
            if (isalpha(*c)) {
                type = 1;
            }
            else if (isdigit(*c)) {
                type = 2;
            }
            else {
                return 0;
            }
        }
        else if (*c == '.') {
            type = 0; 
        }
        else if (type == 1) {
            if (!(isalpha(*c) || isdigit(*c) || *c == '_')) {
                return 0;
            }
        }
        else {
            if (!isdigit(*c)) {
                return 0;
            }
        }
        c++;
    }
    return type != 0;
}

/*
 * Function: Assignment
 * --------------------
 * Assign expression to variable.
 *
 * dst: variable
 */
void Assignment(Variable *dst) {
    GetNext(0);
    if (sNextType == TOK_RESERVED && sNextReserved == RES_FUNCTION) {
        /* Function already built? Its index + 1 is stored in the token's
           extra field. */
        if (sToken->extra) {
            Var_SetFunction(dst, sToken->extra - 1);
            Search(')');
            GetNext(0);
            Expect('{');
            SkipBlock();
        }
        else {
            Var_SetFunction(dst, BuildFunction());
            Expect('{');
            SkipBlock();
        }
    }
    else {
        /* Expression. */
        Expression();
        Var_Set(dst, &sRegister[0]);
        Var_Free(&sRegister[0]);
    }
}

/*
 * Function: DeepAssignment
 * ------------------------
 * Deep copy variable.
 *
 * dst: variable
 */
void DeepAssignment(Variable *dst) {
    Variable *src;
    GetNext(0);
    if (sNextType != TOK_NAME) {
        sprintf(sErrorMsg, "Expected variable");
        Error();
    }
    src = MM_GetVar(sNextKey);
    GetNext(0);
    if (sNextType == TOK_CHARACTER && sNextChar == '(') {
        CallFunc(src);
        MM_CopyVar(dst, MM_GetVar("_"));
        MM_DeleteVar("_");
    }
    else {
        MM_CopyVar(dst, src);
    }
}

/*
 * Function: CallFunc
 * ------------------
 */
void CallFunc(Variable *var) {
    int oldLoopScope = sLoopScope;
    int paramIndex = 0;
    Token *token;
    Function *func;
    Variable *selfVar;
    Variable *thisVar;
    char name[256]; /* ? */

    GetNext(0);

    if (!(var && var->type == VAR_FUNCTION)) {
        sprintf(sErrorMsg, "Variable is not a function");
        Error();
    }
    func = &sFunctions[var->value.i];

    sLoopScope = 0;

    /* self is actually the parent, used for object based stuff. */
    sprintf(name, "%d.self", sScope + 1);
    selfVar = MM_CreateVar(name);
    if (var->parent) {
        selfVar->type = VAR_ALIAS;
        selfVar->value.a = var->parent;
        Var__AddReference(var->parent, selfVar);        
    }
    /* this is the function itself. But this can of course have sub-
       variables. */
    sprintf(name, "%d.this", sScope + 1);
    thisVar = MM_CreateVar(name);
    thisVar->type = VAR_ALIAS;
    thisVar->value.a = var;
    Var__AddReference(var, thisVar);
  
    do {
        Variable *param;

        if (sNextType == TOK_CHARACTER && sNextChar == ')') break;
        if (paramIndex >= func->paramCount) {
            sprintf(sErrorMsg, "Too many parameters in function call");
            Error();
        }
        if (func->params[paramIndex] == 'r') {
            if (sNextType != TOK_NAME) {
                sprintf(sErrorMsg, "Expected variable as argument");
                Error();
            }
            Var_SetAlias(&sRegister[0], sNextKey);
            GetNext(0);
        }
        else {
            Expression();
        }
        sprintf(name, "%d.%s", sScope + 1, func->paramNames[paramIndex]);
        param = MM_CreateVar(name);
        Var_Set(param, &sRegister[0]);
        Var_Free(&sRegister[0]);

        paramIndex++;
        
        if (!(sNextType == TOK_CHARACTER && sNextChar == ')')) {
            Expect(',');
        }
    } while (1);
    token = sToken;   

    sToken = func->token;
    sScope++;
    GetNext(0);
    ExecuteBlock();
    sprintf(name, "%d", sScope);
    MM_DeleteVar(name);
    sScope--;

    sToken = token;
    GetNext(0); /* ) */
    sLoopScope = oldLoopScope;
}

/*
 * Function: Return
 * ----------------
 */
void Return() {
    /*Variable *ret;

    if (sScope <= 0) {
        sprintf(sErrorMsg, "Unexpected return");
        Error();
    }
    GetNext(0);

    MM_DeleteVar("_");
    ret = MM_CreateVar("_");
    if (sNextType == TOK_CHARACTER && sNextChar == ';') {
        Var_Set(ret, 0);
    }
    else if (sNextType == TOK_RESERVED && sNextReserved == RES_ALIAS) {
        Variable *var;
        GetNext(0);
        if (sNextType != TOK_NAME) {
            sprintf(sErrorMsg, "Expected variable");
            Error();
        }
        var = MM_GetVar(sNextKey);
        MM_CopyVar(ret, var);
        GetNext(0);
    }
    else {
        Expression();
        Var_Set(ret, &sRegister[0]);   
    }
    EndStatement();*/

    Variable *ret;

    if (sScope <= 0) {
        sprintf(sErrorMsg, "Unexpected return");
        Error();
    }
    GetNext(0);

    if (sNextType == TOK_CHARACTER && sNextChar == ';') {
        ret = MM_CreateVar("_");
        Var_Set(ret, 0);
    }
    else if (sNextType == TOK_RESERVED && sNextReserved == RES_ALIAS) {
        Variable *var;
        GetNext(0);
        if (sNextType != TOK_NAME) {
            sprintf(sErrorMsg, "Expected variable");
            Error();
        }
        var = MM_GetVar(sNextKey);
        ret = MM_CreateVar("_");
        MM_CopyVar(ret, var);
        GetNext(0);
    }
    else {
        Expression();
        ret = MM_CreateVar("_");
        Var_Set(ret, &sRegister[0]);   
    }
    EndStatement();
}


/*
 * Function: IncVariable
 * ---------------------
 * Increase variable with expression (var +=).
 *
 * var: variable
 */
void IncVariable(Variable *var) {
    /* CreateVar is called before, so var definitely exists. */
    if (var->type == VAR_NUMBER) {
        GetNext(0);
        Expression();
        if (sRegister[0].type == VAR_NUMBER) {
            var->value.d += sRegister[0].value.d;
        }
        else if (sRegister[0].type == VAR_STRING) {
            sprintf(sErrorMsg, "+= is undefined between numbers and strings");
            Error();
        }
        else if (sRegister[0].type == VAR_FUNCTION) {
            sprintf(sErrorMsg, "+= is undefined between numbers and functions");
            Error();
        }
        else {
            sprintf(sErrorMsg, "+= is undefined between numbers and undefined");
            Error();
        }
    }
    else if (var->type == VAR_STRING) {
        GetNext(0);
        Expression();
        if (sRegister[0].type == VAR_STRING) {
            strcat(var->value.s, sRegister[0].value.s);
            Var_Free(&sRegister[0]);
        }
        else if (sRegister[0].type == VAR_NUMBER) {
            sprintf(sErrorMsg, "+= is undefined between strings and numbers");
            Error();
        }
        else if (sRegister[0].type == VAR_FUNCTION) {
            sprintf(sErrorMsg, "+= is undefined between strings and functions");
            Error();
        }
        else {
            sprintf(sErrorMsg, "+= is undefined between strings and undefined");
            Error();
        }
    }
    else if (var->type == VAR_FUNCTION) {
        sprintf(sErrorMsg, "+= is undefined for functions");
        Error();
    }
    else {
        sprintf(sErrorMsg, "+= is undefined for undefined");
        Error();
    }
}

/*
 * Function: DecVariable
 * ---------------------
 * Decrease variable with expression (var -=).
 *
 * var: variable
 */
void DecVariable(Variable *var) {
    /* CreateVar is called before, so var definitely exists. */
    if (var->type == VAR_NUMBER) {
        GetNext(0);
        Expression();
        if (sRegister[0].type == VAR_NUMBER) {
            var->value.d -= sRegister[0].value.d;
        }
        else if (sRegister[0].type == VAR_STRING) {
            sprintf(sErrorMsg, "-= is undefined between numbers and strings");
            Error();
        }
        else if (sRegister[0].type == VAR_FUNCTION) {
            sprintf(sErrorMsg, "-= is undefined between numbers and functions");
            Error();
        }
        else {
            sprintf(sErrorMsg, "-= is undefined between numbers and undefined");
            Error();
        }
    }
    else if (var->type == VAR_STRING) {
        sprintf(sErrorMsg, "-= is undefined for strings");
        Error();
    }
    else if (var->type == VAR_FUNCTION) {
        sprintf(sErrorMsg, "-= is undefined for functions");
        Error();
    }
    else {
        sprintf(sErrorMsg, "-= is undefined for undefined");
        Error();
    }
}

/*
 * Function: MulVariable
 * ---------------------
 * Multiply variable with expression (var *=).
 *
 * var: variable
 */
void MulVariable(Variable *var) {
    /* CreateVar is called before, so var definitely exists. */
    if (var->type == VAR_NUMBER) {
        GetNext(0);
        Expression();
        if (sRegister[0].type == VAR_NUMBER) {
            var->value.d *= sRegister[0].value.d;
        }
        else if (sRegister[0].type == VAR_STRING) {
            sprintf(sErrorMsg, "*= is undefined between numbers and strings");
            Error();
        }
        else if (sRegister[0].type == VAR_FUNCTION) {
            sprintf(sErrorMsg, "*= is undefined between numbers and functions");
            Error();
        }
        else {
            sprintf(sErrorMsg, "*= is undefined between numbers and undefined");
            Error();
        }
    }
    else if (var->type == VAR_STRING) {
        sprintf(sErrorMsg, "*= is undefined for strings");
        Error();
    }
    else if (var->type == VAR_FUNCTION) {
        sprintf(sErrorMsg, "*= is undefined for functions");
        Error();
    }
    else {
        sprintf(sErrorMsg, "*= is undefined for undefined");
        Error();
    }
}

/*
 * Function: DivVariable
 * ---------------------
 * Divide variable with expression (var /=).
 *
 * var: variable
 */
void DivVariable(Variable *var) {
    /* CreateVar is called before, so var definitely exists. */
    if (var->type == VAR_NUMBER) {
        GetNext(0);
        Expression();
        if (sRegister[0].type == VAR_NUMBER) {
            var->value.d /= sRegister[0].value.d;
        }
        else if (sRegister[0].type == VAR_STRING) {
            sprintf(sErrorMsg, "/= is undefined between numbers and strings");
            Error();
        }
        else if (sRegister[0].type == VAR_FUNCTION) {
            sprintf(sErrorMsg, "/= is undefined between numbers and functions");
            Error();
        }
        else {
            sprintf(sErrorMsg, "/= is undefined between numbers and undefined");
            Error();
        }
    }
    else if (var->type == VAR_STRING) {
        sprintf(sErrorMsg, "/= is undefined for strings");
        Error();
    }
    else if (var->type == VAR_FUNCTION) {
        sprintf(sErrorMsg, "/= is undefined for functions");
        Error();
    }
    else {
        sprintf(sErrorMsg, "/= is undefined for undefined");
        Error();
    }
}

/*
 * Function: VariableInc
 * ---------------------
 * Increase variable by 1 (++var).
 *
 * var: variable
 */
void VariableInc(Variable *var) {
    /* Can var be 0? */
    if (var->type == VAR_NUMBER) {
        var->value.d += 1;
    }
    else if (var->type == VAR_STRING) {
        sprintf(sErrorMsg, "++ is undefined for strings");
        Error();
    }
    else if (var->type == VAR_FUNCTION) {
        sprintf(sErrorMsg, "++ is undefined for functions");
        Error();
    }
    else {
        sprintf(sErrorMsg, "++ is undefined for undefined");
        Error();
    }
}

/*
 * Function: VariableDec
 * ---------------------
 * Decrease variable by 1 (--var).
 *
 * var: variable
 */
void VariableDec(Variable *var) {
    /* Can var be 0? */
    if (var->type == VAR_NUMBER) {
        var->value.d -= 1;
    }
    else if (var->type == VAR_STRING) {
        sprintf(sErrorMsg, "-- is undefined for strings");
        Error();
    }
    else if (var->type == VAR_FUNCTION) {
        sprintf(sErrorMsg, "-- is undefined for functions");
        Error();
    }
    else {
        sprintf(sErrorMsg, "-- is undefined for undefined");
        Error();
    }
}

/*
 * Function: Expression
 * --------------------
 * Evaluate expression and put result in register 0.
 */ 
void Expression() {
    VarType type;

    Term();
    type = sRegister[0].type;
    while (sNextType == TOK_CHARACTER && (sNextChar == '+' || sNextChar == '-')) {
        Push(&sRegister[0]);
        if (sNextChar == '+') {
            if (type == VAR_NUMBER) {
                Add();
            }
            else if (type == VAR_STRING) {
                AddString();
            }
            else if (type == VAR_FUNCTION) {
                sprintf(sErrorMsg, "+ is undefined for functions");
                Error();
            }
            else if (type == VAR_UNDEFINED) {
                sprintf(sErrorMsg, "+ is undefined for undefined");
                Error();
            }
            else {
                sprintf(sErrorMsg, "Syntax error");
                Error();
            }
        }
        else {
            if (type == VAR_NUMBER) {
                Sub();
            }
            else if (type == VAR_STRING) {
                sprintf(sErrorMsg, "- is undefined for strings");
                Error();
            }
            else if (type == VAR_FUNCTION) {
                sprintf(sErrorMsg, "- is undefined for functions");
                Error();
            }
            else if (type == VAR_UNDEFINED) {
                sprintf(sErrorMsg, "- is undefined for undefined");
                Error();
            }
            else {
                sprintf(sErrorMsg, "Syntax error");
                Error();
            }
        }
    }
}

/*
 * Function: Term
 * --------------
 */
void Term() {
    SignedFactor();
    while (sNextType == TOK_CHARACTER && (sNextChar == '*' || sNextChar == '/')) {
        Push(&sRegister[0]);
        if (sNextChar == '*') {
            Mul();
        }
        else {
            Div();
        }
    }
}

/*
 * Function: SignedFactor
 * ----------------------
 */
void SignedFactor() {
    if (sNextType == TOK_CHARACTER) {
        if (sNextChar == '+') {
            VarType type;
            GetNext(0);
            Factor();
            type = sRegister[0].type;
            if (type == VAR_NUMBER || type == VAR_STRING) {
            }
            else if (type == VAR_FUNCTION) {
                sprintf(sErrorMsg, "+ is undefined for functions");
                Error();
            }
            else if (type == VAR_UNDEFINED) {
                sprintf(sErrorMsg, "+ is undefined for undefined");
                Error();
            }
            else {
                sprintf(sErrorMsg, "Syntax error");
                Error();
            }
            
        }
        else if (sNextChar == '-') {
            VarType type;
            GetNext(0);
            Factor();
            type = sRegister[0].type;
            if (type == VAR_NUMBER) {
                sRegister[0].value.d = -sRegister[0].value.d;
            }
            else {
                if (type == VAR_STRING) {
                    sprintf(sErrorMsg, "- is undefined for strings");
                    Error();
                }
                else if (type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, "- is undefined for functions");
                    Error();
                }
                else if (type == VAR_UNDEFINED) {
                    sprintf(sErrorMsg, "- is undefined for undefined");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Syntax error");
                    Error();
                }
            }
        }
        else {
            Factor();
        }
    }
    else {
        Factor();
    }
}

/*
 * Function: Factor
 * ----------------
 */
void Factor() {
    if (sNextType == TOK_CHARACTER) {
        if (sNextChar == '(') {
            GetNext(0);
            Expression();
            Expect(')');
        }
        else if (sNextChar == CHR_INC) {
            Variable *var;
            GetNext(0);
            if (sNextType == TOK_NAME) {
                var = MM_GetVar(sNextKey);
                if (var) {
                    if (var->type == VAR_NUMBER) {
                        var->value.d += 1;
                        Var_Set(&sRegister[0], var);
                        GetNext(0);
                        CheckPostMod(var);
                    }
                    else if (var->type == VAR_STRING) {
                        sprintf(sErrorMsg, "++ is undefined for strings");
                        Error();
                    }
                    else if (var->type == VAR_FUNCTION) {
                        sprintf(sErrorMsg, "++ is undefined for functions");
                        Error();
                    }
                    else if (var->type == VAR_UNDEFINED) {
                        sprintf(sErrorMsg, "++ is undefined for undefined");
                        Error();
                    }
                    else {
                    }
                }
                else {
                    sprintf(sErrorMsg, "++ is undefined for undefined");
                    Error();
                }
            }
            else {
                sprintf(sErrorMsg, "Expected variable");
                Error();
            }
        }
        else if (sNextChar == CHR_DEC) {
            Variable *var;
            GetNext(0);
            if (sNextType == TOK_NAME) {
                var = MM_GetVar(sNextKey);
                if (var) {
                    if (var->type == VAR_NUMBER) {
                        var->value.d -= 1;
                        Var_Set(&sRegister[0], var);
                        GetNext(0);
                        CheckPostMod(var);
                    }
                    else if (var->type == VAR_STRING) {
                        sprintf(sErrorMsg, "-- is undefined for strings");
                        Error();
                    }
                    else if (var->type == VAR_FUNCTION) {
                        sprintf(sErrorMsg, "-- is undefined for functions");
                        Error();
                    }
                    else if (var->type == VAR_UNDEFINED) {
                        sprintf(sErrorMsg, "-- is undefined for undefined");
                        Error();
                    }
                }
                else {
                    sprintf(sErrorMsg, "-- is undefined for undefined");
                    Error();
                }
            }
            else {
                sprintf(sErrorMsg, "Expected variable");
                Error();
            }
        }
        else {
            sprintf(sErrorMsg, "Unexpected character");
            Error();
        }
    }
    else if (sNextType == TOK_NAME) {
        Variable *var = MM_GetVar(sNextKey);
        if (var) {
            VarType type = var->type;
            if (type == VAR_NUMBER) {
                GetNext(0);
                Var_Set(&sRegister[0], var);
                CheckPostMod(var);
            }
            else if (type == VAR_STRING) {
                GetNext(0);
                Var_Set(&sRegister[0], var);
            }
            else if (type == VAR_FUNCTION) {
                GetNext(0);
                if (sNextType == TOK_CHARACTER && sNextChar == '(') {
                    CallFunc(var);
                    Var_Set(&sRegister[0], MM_GetVar("_"));
                    MM_DeleteVar("_");
                }
                else {
                    Var_Set(&sRegister[0], var);
                }
            }
            else if (type == VAR_UNDEFINED) {
                GetNext(0);
                Var_Set(&sRegister[0], var);
            }
        }
        else {
            GetNext(0);
            /* Will be set do undefined. */
            Var_Set(&sRegister[0], 0);
        }
    }
    else if (sNextType == TOK_NUMBER) {
        sRegister[0].type = VAR_NUMBER;
        sRegister[0].value.d = sNextNumber;
        GetNext(0);
    }
    else if (sNextType == TOK_STRING) {
        sRegister[0].type = VAR_STRING;
        sRegister[0].value.s = strdup(sNextString);
        GetNext(0);
    }
    else if (sNextType == TOK_RESERVED) {
        switch (sNextReserved) {
        case RES_TOSTRING:
            ToString();
            break;
        case RES_TONUMBER:
            ToNumber();
            break;
        case RES_SIZEOF:
            SizeOf();
            break;
        case RES_TIME:
            Time();
            break;
        default:
            sprintf(sErrorMsg, "Syntax error");
            Error();
            break;
        }

    }
    else {
        sprintf(sErrorMsg, "Expected expression");
        Error();
    }
}

/*
 * Function: Add
 * -------------
 */
void Add() {
    VarType type;
    GetNext(0);
    Term();
    type = sRegister[0].type;
    if (type == VAR_NUMBER) {
        Variable var;
        Pop(&var);
        sRegister[0].value.d += var.value.d;
    }
    else if (type == VAR_STRING) {
        sprintf(sErrorMsg, "+ is undefined between numbers and strings");
        Error();
    }
    else if (type == VAR_FUNCTION) {
        sprintf(sErrorMsg, "+ is undefined between numbers an functions");
        Error();
    }
    else if (type == VAR_UNDEFINED) {
        sprintf(sErrorMsg, "+ is undefined between numbers and undefined");
        Error();
    }
    else {
        INTERNAL_ERROR;
    }
}

/*
 * Function: Sub
 * -------------
 */
void Sub() {
    VarType type;
    GetNext(0);
    Term();
    type = sRegister[0].type;
    if (type == VAR_NUMBER) {
        Variable var;
        Pop(&var);
        sRegister[0].value.d -= var.value.d;
        sRegister[0].value.d = -sRegister[0].value.d;
    }
    else if (type == VAR_STRING) {
        sprintf(sErrorMsg, "- is undefined between numbers and strings");
        Error();
    }
    else if (type == VAR_FUNCTION) {
        sprintf(sErrorMsg, "- is undefined between numbers and functions");
        Error();
    }
    else if (type == VAR_UNDEFINED) {
        sprintf(sErrorMsg, "- is undefined between numbers and undefined");
        Error();
    }
    else {
        INTERNAL_ERROR;
    }
}

/*
 * Function: Mul
 * -------------
 */
void Mul() {
    VarType type = sStack[sStackIndex - 1].type;
    GetNext(0);
    if (type == VAR_NUMBER) {
        SignedFactor();
        type = sRegister[0].type;
        if (type == VAR_NUMBER) {
            Variable var;
            Pop(&var);
            sRegister[0].value.d *= var.value.d;
        }
        else if (type == VAR_STRING) {
            sprintf(sErrorMsg, "* is undefined between numbers and strings");
            Error();
        }
        else if (type == VAR_FUNCTION) {
            sprintf(sErrorMsg, "* is undefined between numbers and functions");
            Error();
        }
        else if (type == VAR_UNDEFINED) {
            sprintf(sErrorMsg, "* is undefined between numbers and undefined");
            Error();
        }
        else {
            INTERNAL_ERROR;
        }
    }
    else if (type == VAR_STRING) {
        sprintf(sErrorMsg, "* is undefined for strings");
        Error();
    }
    else if (type == VAR_FUNCTION) {
        sprintf(sErrorMsg, "* is undefined for functions");
        Error();
    }
    else if (type == VAR_UNDEFINED) {
        sprintf(sErrorMsg, "* is undefined for undefined");
        Error();
    }
    else {
        INTERNAL_ERROR;
    }
}

/*
 * Function: Div
 * -------------
 */
void Div() {
    VarType type = sStack[sStackIndex - 1].type;
    GetNext(0);
    if (type == VAR_NUMBER) {
        SignedFactor();
        type = sRegister[0].type;
        if (type == VAR_NUMBER) {
            Pop(&sRegister[1]);
            sRegister[1].value.d /= sRegister[0].value.d;
            sRegister[0] = sRegister[1]; 
        }
        else if (type == VAR_STRING) {
            sprintf(sErrorMsg, "/ is undefined between numbers and strings");
            Error();
        }
        else if (type == VAR_FUNCTION) {
            sprintf(sErrorMsg, "/ is undefined between numbers and functions");
            Error();
        }
        else if (type == VAR_UNDEFINED) {
            sprintf(sErrorMsg, "/ is undefined between numbers and undefined");
            Error();
        }
        else {
            INTERNAL_ERROR;
        }
    }
    else if (type == VAR_STRING) {
        sprintf(sErrorMsg, "/ is undefined for strings");
        Error();
    }
    else if (type == VAR_FUNCTION) {
        sprintf(sErrorMsg, "/ is undefined for functions");
        Error();
    }
    else if (type == VAR_UNDEFINED) {
        sprintf(sErrorMsg, "/ is undefined for undefined");
        Error();
    }
    else {
        INTERNAL_ERROR;
    }
}

/*
 * Function: AddString
 * -------------------
 */
void AddString() {
    VarType type;
    GetNext(0);
    Term();
    type = sRegister[0].type;
    if (type == VAR_STRING) {
        Variable var;
        size_t len;
        char *newString;
        Pop(&var);
        len = strlen(sRegister[0].value.s) + strlen(var.value.s);
        newString = (char *)malloc(sizeof(char)*(len + 1));
        strcpy(newString, var.value.s);
        strcat(newString, sRegister[0].value.s);
        free(var.value.s);
        free(sRegister[0].value.s);
        sRegister[0].value.s = newString;
    }
    else if (type == VAR_NUMBER) {
        sprintf(sErrorMsg, "+ is undefined between strings and numbers");
        Error();
    }
    else if (type == VAR_FUNCTION) {
        sprintf(sErrorMsg, "+ is undefined between strings and functions");
        Error();
    }
    else if (type == VAR_UNDEFINED) {
        sprintf(sErrorMsg, "+ is undefined between strings and undefined");
        Error();
    }
    else {
        INTERNAL_ERROR;
    }   
}

/*
 * Function: CheckPostMod
 * ----------------------
 * Apply any post modifiers (++ and --) on variable.
 *
 * var: variable
 */
void CheckPostMod(Variable *var) {
    if (sNextType == TOK_CHARACTER) {
        if (sNextChar == CHR_INC) {
            if (var) {
                if (var->type == VAR_NUMBER) {
                    var->value.d += 1;
                    GetNext(0);
                }
                else if (var->type == VAR_STRING) {
                    sprintf(sErrorMsg, "++ is undefined for strings");
                    Error();
                }
                else if (var->type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, "++ is undefined for functions");
                    Error();
                }
                else if (var->type == VAR_UNDEFINED) {
                    sprintf(sErrorMsg, "++ is undefined for undefined");
                    Error();
                }
                else {
                    INTERNAL_ERROR;
                }
            }
            else {
                sprintf(sErrorMsg, "++ is undefined for undefined");
                Error();
            }
        }
        else if (sNextChar == CHR_DEC) {
            if (var) {
                if (var->type == VAR_NUMBER) {
                    var->value.d -= 1;
                    GetNext(0);
                }
                else if (var->type == VAR_STRING) {
                    sprintf(sErrorMsg, "-- is undefined for strings");
                    Error();
                }
                else if (var->type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, "-- is undefined for functions");
                    Error();
                }
                else if (var->type == VAR_UNDEFINED) {
                    sprintf(sErrorMsg, "-- is undefined for undefined");
                    Error();
                }
                else {
                    INTERNAL_ERROR;
                }
            }
            else {
                sprintf(sErrorMsg, "-- is undefined for undefined");
                Error();
            }
        }
    }
}

/*
 * Function: BoolExpression
 * ------------------------
 */
void BoolExpression() {
    SingleBool();
    while (sNextType == TOK_RESERVED && (sNextReserved == RES_AND || sNextReserved == RES_OR)) {
        if (sNextReserved == RES_AND) {
            /* Short circuit evaluateion, skip to end of this expression. This
               could probably be done by just increasing sToken instead. */
            if (sRegister[0].value.i == 0) {
                int parc = 1;
                GetNext(0);
                do {
                    if (sNextType == TOK_CHARACTER) {
                        if (sNextChar == '(') {
                            parc++;
                        }
                        else if (sNextChar == ')') {
                            if (--parc == 0) break;
                        }
                        else if (sNextChar == ';') {
                            sprintf(sErrorMsg, "Expected )");
                            Error();
                        }
                    }
                    else if (sNextType == TOK_EOF) {
                        sprintf(sErrorMsg, "Expected )");
                        Error();
                    }
                    GetNext(0);
                } while (1);
            }
            else {
                Push(&sRegister[0]);
                GetNext(0);
                SingleBool();
                sRegister[0].value.i &= sStack[--sStackIndex].value.i;
            }
        }
        /* RES_OR */
        else {
            /* Short circuit evaluation, skip to the end of this expression. */
            if (sRegister[0].value.i == 1) {
                int parc = 1;
                GetNext(0);
                do {
                    if (sNextType == TOK_CHARACTER) {
                        if (sNextChar == '(') {
                            parc++;
                        }
                        else if (sNextChar == ')') {
                            if (--parc == 0) break;
                        }
                        else if (sNextChar == ';') {
                            sprintf(sErrorMsg, "Expected )");
                            Error();
                        }
                    }
                    else if (sNextType == EOF) {
                        sprintf(sErrorMsg, "Expected )");
                        Error();
                    }
                    GetNext(0);
                } while (1);
            }
            else {
                Push(&sRegister[0]);
                GetNext(0);
                SingleBool();
                sRegister[0].value.i |= sStack[--sStackIndex].value.i;
            }
        }
    }
}

/*
 * Function: SingleBool
 * --------------------
 */
void SingleBool() {
    if (sNextType == TOK_RESERVED && sNextReserved == RES_NOT) {
        GetNext(0);
        BoolExpression();
        sRegister[0].value.i = !sRegister[0].value.i;
    }
    else if (sNextType == TOK_RESERVED && sNextReserved == RES_DEFINED) {
        GetNext(0);
        Expect('(');
        if (sNextType != TOK_NAME) {
            sprintf(sErrorMsg, "Expected name");
            Error();
        }
        sRegister[0].value.i = MM_VarDefined(sNextKey);
        GetNext(0);
        Expect(')');
    }
    else if (sNextType == TOK_CHARACTER && sNextChar == '(') {
        Token *check = sToken;
        int parc = 1;
        while (parc > 0) {
            if (check->type == TOK_EOF) {
                sprintf(sErrorMsg, "Expected )");
                Error();
            }
            else if (check->type == TOK_CHARACTER) {
                if (check->value.i == '(') {
                    parc++;
                }
                else if (check->value.i == ')') {
                    parc--;
                }
            }
            check++;
        }
        if (check->type == TOK_CHARACTER && IsRelOp(check->value.i)) {
            /* Next is still ( character. */
            BoolTerm();
        }
        else {
            GetNext(0);
            BoolExpression();
            Expect(')');
        }
    }
    else {
        BoolTerm();
    }
}

/*
 * Function: BoolTerm
 * ------------------
 */
void BoolTerm() {
    int type;
    Expression();
    type = sRegister[0].type;
    if (sNextType == TOK_CHARACTER && IsRelOp(sNextChar)) {
        if (type == VAR_NUMBER) {
            switch (sNextChar) { 
            case '=':
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_NUMBER) {
                    sRegister[0].value.i = sStack[--sStackIndex].value.d == sRegister[0].value.d;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    Pop(&sRegister[0]);
                    sRegister[0].value.i = 0;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_STRING) {
                    sprintf(sErrorMsg, "= is undefined between numbers and strings");
                    Error();
                }
                else if (sRegister[0].type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, "= is undefined between numbers and functions");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Expected numeric expression");
                    Error();
                }
                break;
            case CHR_NOTEQ:
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_NUMBER) {
                    sRegister[0].value.i = sStack[--sStackIndex].value.d != sRegister[0].value.d;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    Pop(&sRegister[0]);
                    sRegister[0].value.i = 1;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_STRING) {
                    sprintf(sErrorMsg, "<> is undefined between numbers and strings");
                    Error();
                }
                else if (sRegister[0].type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, "<> is undefined between numbers and functions");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Expected numeric expression");
                    Error();
                }
                break;
            case '<':
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_NUMBER) {
                    sRegister[0].value.i = sStack[--sStackIndex].value.d < sRegister[0].value.d;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_STRING) {
                    sprintf(sErrorMsg, "< is undefined between numbers and strings");
                    Error();
                }
                else if (sRegister[0].type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, "< is undefined between numbers and functions");
                    Error();
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    sprintf(sErrorMsg, "< is undefined between numbers and undefined");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Expected numeric expression");
                    Error();
                }
                break;
            case CHR_LESSOREQ:
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_NUMBER) {
                    sRegister[0].value.i = sStack[--sStackIndex].value.d <= sRegister[0].value.d;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_STRING) {
                    sprintf(sErrorMsg, "<= is undefined between numbers and strings");
                    Error();
                }
                else if (sRegister[0].type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, "<= is undefined between numbers and functions");
                    Error();
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    sprintf(sErrorMsg, "<= is undefined between numbers and undefined");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Expected numeric expression");
                    Error();
                }
                break;
            case '>':
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_NUMBER) {
                    sRegister[0].value.i = sStack[--sStackIndex].value.d > sRegister[0].value.d;
                    sRegister[0].type = VAR_UNDEFINED;                    
                }
                else if (sRegister[0].type == VAR_STRING) {
                    sprintf(sErrorMsg, "> is undefined between numbers and strings");
                    Error();
                }
                else if (sRegister[0].type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, "> is undefined between numbers and functions");
                    Error();
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    sprintf(sErrorMsg, "> is undefined between numbers and undefined");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Expected numeric expression");
                    Error();
                }
                break;
            case CHR_GREATEROREQ:
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_NUMBER) {
                    sRegister[0].value.i = sStack[--sStackIndex].value.d >= sRegister[0].value.d;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_STRING) {
                    sprintf(sErrorMsg, ">= is undefined between numbers and strings");
                    Error();
                }
                else if (sRegister[0].type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, ">= is undefined between numbers and functions");
                    Error();
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    sprintf(sErrorMsg, ">= is undefined between numbers and undefined");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Expected numeric expression");
                    Error();
                }
                break;
            default:
                INTERNAL_ERROR;
                break;
            }
        }
        else if (type == VAR_STRING) {
            switch (sNextChar) {
            case '=':
                GetNext(0);
                /* Push set's type of sRegister[0] to undefined, so expression
                   does not free the string. */
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_STRING) {
                    int cmp = strcmp(sStack[--sStackIndex].value.s, sRegister[0].value.s); 
                    Var_Free(&sRegister[0]);
                    Var_Free(&sStack[sStackIndex]);
                    sRegister[0].value.i = cmp == 0;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    Pop(&sRegister[0]);
                    Var_Free(&sRegister[0]);
                    sRegister[0].value.i = 0;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_NUMBER) {
                    sprintf(sErrorMsg, "= is undefined between strings and numbers");
                    Error();
                }
                else if (sRegister[0].type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, "= is undefined between strings and functions");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Expected string expression");
                    Error();
                }
                break;
            case CHR_NOTEQ:
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_STRING) {
                    int cmp = strcmp(sStack[--sStackIndex].value.s, sRegister[0].value.s); 
                    Var_Free(&sRegister[0]);
                    Var_Free(&sStack[sStackIndex]);
                    sRegister[0].value.i = cmp != 0;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    Pop(&sRegister[0]);
                    Var_Free(&sRegister[0]);
                    sRegister[0].value.i = 1;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_NUMBER) {
                    sprintf(sErrorMsg, "<> is undefined between strings and numbers");
                    Error();
                }
                else if (sRegister[0].type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, "<> is undefined between strings and functions");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Expected string expression");
                    Error();
                }
                break;
            case '<':
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_STRING) {
                    int res = strlen(sStack[--sStackIndex].value.s) < strlen(sRegister[0].value.s); 
                    Var_Free(&sRegister[0]);
                    Var_Free(&sStack[sStackIndex]);
                    sRegister[0].value.i = res;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_NUMBER) {
                    sprintf(sErrorMsg, "< is undefined between strings and numbers");
                    Error();
                }
                else if (sRegister[0].type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, "< is undefined between strings and functions");
                    Error();
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    sprintf(sErrorMsg, "< is undefined between strings and undefined");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Expected string expression");
                    Error();
                }
                break;
            case CHR_LESSOREQ:
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_STRING) {
                    int res = strlen(sStack[--sStackIndex].value.s) <= strlen(sRegister[0].value.s); 
                    Var_Free(&sRegister[0]);
                    Var_Free(&sStack[sStackIndex]);
                    sRegister[0].value.i = res;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_NUMBER) {
                    sprintf(sErrorMsg, "<= is undefined between strings and numbers");
                    Error();
                }
                else if (sRegister[0].type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, "<= is undefined between strings and functions");
                    Error();
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    sprintf(sErrorMsg, "<= is undefined between strings and undefined");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Expected string expression");
                    Error();
                }
                break;
            case '>':
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_STRING) {
                    int res = strlen(sStack[--sStackIndex].value.s) > strlen(sRegister[0].value.s); 
                    Var_Free(&sRegister[0]);
                    Var_Free(&sStack[sStackIndex]);
                    sRegister[0].value.i = res;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_NUMBER) {
                    sprintf(sErrorMsg, "> is undefined between strings and numbers");
                    Error();
                }
                else if (sRegister[0].type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, "> is undefined between strings and functions");
                    Error();
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    sprintf(sErrorMsg, "> is undefined between strings and undefined");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Expected string expression");
                    Error();
                }

                break;
            case CHR_GREATEROREQ:
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_STRING) {
                    int res = strlen(sStack[--sStackIndex].value.s) >= strlen(sRegister[0].value.s); 
                    Var_Free(&sRegister[0]);
                    Var_Free(&sStack[sStackIndex]);
                    sRegister[0].value.i = res;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_NUMBER) {
                    sprintf(sErrorMsg, ">= is undefined between strings and numbers");
                    Error();
                }
                else if (sRegister[0].type == VAR_FUNCTION) {
                    sprintf(sErrorMsg, ">= is undefined between strings and functions");
                    Error();
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    sprintf(sErrorMsg, ">= is undefined between strings and undefined");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Expected string expression");
                    Error();
                }
                break;
            default:
                INTERNAL_ERROR;
                break;
            }
        }
        else if (type == VAR_FUNCTION) {
            switch (sNextChar) {
            case '=':
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_FUNCTION) {
                    sRegister[0].value.i = sStack[--sStackIndex].value.i == sRegister[0].value.i;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    Pop(&sRegister[0]);
                    sRegister[0].value.i = 0;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_NUMBER) {
                    sprintf(sErrorMsg, "= is undefined between functions and numbers");
                    Error();
                }
                else if (sRegister[0].type == VAR_STRING) {
                    sprintf(sErrorMsg, "= is undefined between functions and strings");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Expected function expression");
                    Error();
                }
                break;
            case CHR_NOTEQ:
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_FUNCTION) {
                    sRegister[0].value.i = sStack[--sStackIndex].value.i != sRegister[0].value.i;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    Pop(&sRegister[0]);
                    sRegister[0].value.i = 1;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_NUMBER) {
                    sprintf(sErrorMsg, "<> is undefined between functions and numbers");
                    Error();
                }
                else if (sRegister[0].type == VAR_STRING) {
                    sprintf(sErrorMsg, "<> is undefined between functions and strings");
                    Error();
                }
                else {
                    sprintf(sErrorMsg, "Expected function expression");
                    Error();
                }
                break;
            case '<':
                sprintf(sErrorMsg, "< is undefined for functions");
                Error();
                break;
            case CHR_LESSOREQ:
                sprintf(sErrorMsg, "<= is undefined for functions");
                Error();
                break;
            case '>':
                sprintf(sErrorMsg, "> is undefined for functions");
                Error();
                break;
            case CHR_GREATEROREQ:
                sprintf(sErrorMsg, ">= is undefined for functions");
                Error();
                break;
            default:
                INTERNAL_ERROR;
                break;
            }
        }
        /* VAR_UNDEFINED */
        else {
            switch (sNextChar) {
            case '=':
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_NUMBER || sRegister[0].type == VAR_FUNCTION) {
                    Pop(&sRegister[0]);
                    sRegister[0].value.i = 0;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_STRING) {
                    Var_Free(&sRegister[0]);
                    Pop(&sRegister[0]);
                    sRegister[0].value.i = 0;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    Pop(&sRegister[0]);
                    sRegister[0].value.i = 1;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else {
                    INTERNAL_ERROR;
                }
                break;
            case CHR_NOTEQ:
                GetNext(0);
                Push(&sRegister[0]);
                Expression();
                if (sRegister[0].type == VAR_NUMBER || sRegister[0].type == VAR_FUNCTION) {
                    Pop(&sRegister[0]);
                    sRegister[0].value.i = 1;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_STRING) {
                    Var_Free(&sRegister[0]);
                    Pop(&sRegister[0]);
                    sRegister[0].value.i = 1;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else if (sRegister[0].type == VAR_UNDEFINED) {
                    Pop(&sRegister[0]);
                    sRegister[0].value.i = 0;
                    sRegister[0].type = VAR_UNDEFINED;
                }
                else {
                    INTERNAL_ERROR;
                }
                break;
            case '<':
                sprintf(sErrorMsg, "< is undefined for undefined");
                Error();
                break;
            case CHR_LESSOREQ:
                sprintf(sErrorMsg, "<= is undefined for undefined");
                break;
            case '>':
                sprintf(sErrorMsg, "> is undefined for undefined");
                break;
            case CHR_GREATEROREQ:
                sprintf(sErrorMsg, ">= is undefined for undefined");
                break;
            default:
                INTERNAL_ERROR;
                break;
            }
        }
    }
    else {
        if (type == VAR_NUMBER) {
            sRegister[0].value.i = sRegister[0].value.d != 0;
        }
        else if (type == VAR_STRING) {
            /*sRegister[0].value.i = strlen(sRegister[0].value.s) != 0;*/
            sRegister[0].value.i = sRegister[0].value.s[0] != 0;
        }
        else if (type == VAR_FUNCTION) {
            sRegister[0].value.i = 1;
        }
        else {
            sRegister[0].value.i = 0;
        }
    }   
}

/*
 * Function: IsRelOp
 * -----------------
 */
int IsRelOp(int c) {
    switch (c) {
    case '=':
    /*case CHR_EQD:*/
    case '<':
    case '>':
    case CHR_LESSOREQ:
    case CHR_GREATEROREQ:
    case CHR_NOTEQ:
        return 1;
    default:
        return 0;
    }
}

/*
 * Function: Push
 * --------------
 * Push variable to stack.
 *
 * var: variable
 */
void Push(Variable *var) {
    sStack[sStackIndex] = *var;
    sStackIndex++;
    var->type = VAR_UNDEFINED;
}

/*
 * Function: Pop
 * -------------
 * Pop variable from stack.
 *
 * var: variable
 */
void Pop(Variable *var) {
    sStackIndex--;
    *var = sStack[sStackIndex];
    sStack[sStackIndex].type = VAR_UNDEFINED;
}

/*
 * Function: Expect
 * ----------------
 * Error if character is missing.
 */
void Expect(int c) {
    if (sNextType == TOK_CHARACTER && sNextChar == c) {
        GetNext(0);
    }
    else {
        sprintf(sErrorMsg, "Missing %c", c);
        Error();
    }
}

/*
 * Function: EndStatement
 * ----------------------
 * Error if ; is missing.
 */
void EndStatement() {
    if (sNextType == TOK_CHARACTER && sNextChar == ';') {
        GetNext(0);
    }
    else {
        sprintf(sErrorMsg, "Missing ;");
        Error();
    }
}

/*
 * Function: Error
 * ---------------
 * Print error message and exit.
 *
 * msg: error message
 */
void Error() {
    printf("ERROR (%d:%d) %s\n", sLastLine, sLastCol, sErrorMsg);
    exit(EXIT_FAILURE);
}

/*
 * Function: Wln
 * -------------
 * wln(<string expr>);
 */
void Wln() {
    GetNext(0);
    Expect('(');
    if (sNextType == TOK_CHARACTER && sNextChar == ')') {
        GetNext(0);
        EndStatement();
        printf("\n");
    }
    else {
        Expression();
        if (sRegister[0].type != VAR_STRING) {
            sprintf(sErrorMsg, "Expected string expression");
            Error();
        }
        Push(&sRegister[0]);
        Expect(')');
        EndStatement();
        Pop(&sRegister[0]);
        printf("%s\n", sRegister[0].value.s);
        Var_Free(&sRegister[0]);
    }
}

/*
 * Function: Write
 * ---------------
 * write(<string expr>);
 */
void Write() {
    GetNext(0);
    Expect('(');
    Expression();
    if (sRegister[0].type != VAR_STRING) {
        sprintf(sErrorMsg, "Expected string expression");
        Error();
    }
    Push(&sRegister[0]);
    Expect(')');
    EndStatement();
    Pop(&sRegister[0]);
    printf("%s", sRegister[0].value.s);
    Var_Free(&sRegister[0]);
}

/*
 * Function: ToString
 * ------------------
 * tostring(<expr>)
 */
void ToString() {
    int dec = 0;
    GetNext(0);
    Expect('(');
    Expression();
    if (sNextType == TOK_CHARACTER && sNextChar == ',') {
        GetNext(0);
        Push(&sRegister[0]);
        Expression();
        if (sRegister[0].type != VAR_NUMBER) {
            sprintf(sErrorMsg, "Expected numeric expression");
            Error();
        }
        dec = (int)fabs(sRegister[0].value.d);
        if (dec > 16) dec = 16;
        Pop(&sRegister[0]);
    }
    if (sRegister[0].type == VAR_NUMBER) {
        char tmp[64];
        /* Uhm, maybe I'm just stupid, idk ... */
        switch (dec) {
        case 1: sprintf(tmp, "%.1f", sRegister[0].value.d); break;
        case 2: sprintf(tmp, "%.2f", sRegister[0].value.d); break;
        case 3: sprintf(tmp, "%.3f", sRegister[0].value.d); break;
        case 4: sprintf(tmp, "%.4f", sRegister[0].value.d); break;
        case 5: sprintf(tmp, "%.5f", sRegister[0].value.d); break;
        case 6: sprintf(tmp, "%.6f", sRegister[0].value.d); break;
        case 7: sprintf(tmp, "%.7f", sRegister[0].value.d); break;
        case 8: sprintf(tmp, "%.8f", sRegister[0].value.d); break;
        case 9: sprintf(tmp, "%.9f", sRegister[0].value.d); break;
        case 10: sprintf(tmp, "%.10f", sRegister[0].value.d); break;
        case 11: sprintf(tmp, "%.11f", sRegister[0].value.d); break;
        case 12: sprintf(tmp, "%.12f", sRegister[0].value.d); break;
        case 13: sprintf(tmp, "%.13f", sRegister[0].value.d); break;
        case 14: sprintf(tmp, "%.14f", sRegister[0].value.d); break;
        case 15: sprintf(tmp, "%.15f", sRegister[0].value.d); break;
        case 16: sprintf(tmp, "%.16f", sRegister[0].value.d); break;
        default:
            sprintf(tmp, "%.0f", sRegister[0].value.d);
            break;        
        }
        Var_SetString(&sRegister[0], tmp);
    }
    else if (sRegister[0].type == VAR_STRING) {
    }
    else if (sRegister[0].type == VAR_FUNCTION) {
        Var_SetString(&sRegister[0], "(FUNCTION)");
    }
    else {
        Var_SetString(&sRegister[0], "(UNDEFINED)");
    }
    Expect(')');
}

/*
 * Function: ToNumber
 * ------------------
 * tonumber(<expr>)
 */
void ToNumber() {
    GetNext(0);
    Expect('(');
    Expression();
    if (sRegister[0].type == VAR_NUMBER) {
    }
    else if (sRegister[0].type == VAR_STRING) {
        double d = atof(sRegister[0].value.s);
        Var_SetNumber(&sRegister[0], d);
    }
    else if (sRegister[0].type == VAR_FUNCTION) {
        sprintf(sErrorMsg, "Can't convert function to number");
        Error();
    }
    else {
        Var_SetNumber(&sRegister[0], 0);
    }
    Expect(')');
}

/*
 * Function: SizeOf
 * ----------------
 * sizeof(<variable>)
 */
void SizeOf() {
    GetNext(0);
    Expect('(');
    if (sNextType != TOK_NAME) {
        sprintf(sErrorMsg, "Expected variable");
        Error();
    }
    Var_SetNumber(&sRegister[0], MM_VarSubCount(sNextKey));
    GetNext(0);
    Expect(')');
}


/*
 * Function: Time
 * --------------
 * time(), time in milliseconds since system boot.
 */
void Time() {
    GetNext(0);
    Expect('(');
    Expect(')');
    struct timespec ts;
    clock_gettime(CLOCK_MONOTONIC, &ts);
    Var_SetNumber(&sRegister[0], (double)ts.tv_sec*1000 + (double)(ts.tv_nsec/1000000));
}

/*
 * Function: Sleep
 * ---------------
 * sleep(milliseconds);
 */
void Sleep() {
    unsigned long ms;
    GetNext(0);
    Expect('(');
    Expression();
    if (sRegister[0].type != VAR_NUMBER) {
        sprintf(sErrorMsg, "Expected numeric expression");
        Error();
    }
    if (sRegister[0].value.d < 0) {
        ms = 0;
    }
    else {
        ms = (unsigned long)sRegister[0].value.d;
    }
    Expect(')');
    EndStatement();
    struct timespec ts;
    ts.tv_sec = ms/1000;
    ts.tv_nsec = (ms%1000)*1000000;
    nanosleep(&ts, 0);
}

/*
 * Function: Free
 * --------------
 * free(variable);
 */
void Free() {
    GetNext(0);
    if (sNextType != TOK_NAME) {
        sprintf(sErrorMsg, "Expected variable");
        Error();
    }
    MM_DeleteVar(sNextKey);
    GetNext(0);
    EndStatement();
}

/*
 * Function: Extern
 * ----------------
 * extern <var>[, <var> ...];
 */
void Extern() {
    Variable *var;
    char *tmpName;

    if (sScope == 0) {
        sprintf(sErrorMsg, "Extern outside function");
        Error();
    }
    do {
        GetNext(1);
        if (sNextType != TOK_NAME) {
            sprintf(sErrorMsg, "Expected variable");
            Error();
        }
        var = MM_GetVar(sNextKey);
        tmpName = (char *)malloc(sizeof(sNextKey)*(strlen(sNextKey) + 5));
        sprintf(tmpName, "%d.%s", sScope, sNextKey);
        var = MM_CreateVar(tmpName);    
        free(tmpName);
        Var_SetAlias(var, sNextKey);
        GetNext(0);
    } while (sNextType == TOK_CHARACTER && sNextChar == ',');
    EndStatement();
}

/*
 * Function: Unalias
 * -----------------
 * unref <var>;
 */
void Unalias() {
    GetNext(0);
    if (sNextType != TOK_NAME) {
        sprintf(sErrorMsg, "Expected variable");
        Error();
    }
    MM_UnaliasVar(sNextKey);
    GetNext(0);
    EndStatement();
}

