/*
 * File: symbol_table.h
 * --------------------
 * Handles tree structures for storing data through access keys.
 *
 * By: Marcus 2018
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "symbol_table.h"

#ifdef ST_REFC
STable *sCurrentTable = 0;
#endif

/*
 * Function: ST_Init
 * -----------------
 * Return a new symbol table. Use ST_Clear before freeing the returned symbol
 * table.
 *
 * freeDataFunc: function to be called for destroying data added to the table
 *               with ST_AddData. Any time data is removed from the table, this
 *               function will be called.
 *
 * Returns: symbol table
 */
STable *ST_Init(void (*freeDataFunc)(void *), void *(*copyDataFunc)(void *)) {
	STable *st;
	
	st = (STable *)malloc(sizeof(STable));
	st->rootNode = 0;
	st->keyCount = 0;
    st->freeData = freeDataFunc;
    st->copyData = copyDataFunc;
#ifdef ST_REFC
    st->refc = 0;
#endif
	
	return st;
}

/*
 * Function: ST_ClearRec
 * ---------------------
 * Internal.
 */
void ST_ClearRec(STNode *node, void (*freeData)(void *)) {
	if (node == 0) {
        return;
    }
	
	ST_ClearRec(node->sister, freeData);
	ST_ClearRec(node->children, freeData);
	
	if (node->data) {
        if (freeData) freeData(node->data);
    }
		
	free(node);
#ifdef ST_REFC
    sCurrentTable->refc--;
#endif
}

/*
 * Function: ST_Clear
 * ------------------
 * Clear symbol table. You need to manually free the table itself.
 *
 * st: symbol table
 */
void ST_Clear(STable *st) {
#ifdef ST_REFC
    sCurrentTable = st;
#endif
	ST_ClearRec(st->rootNode, st->freeData);
	st->rootNode = 0;
#ifdef ST_REFC
    assert(st->refc == 0);
#endif
}

/*
 * Function: ST_Fill
 * -----------------
 * Internal.
 */
void ST_Fill(STNode *node, const char *key, void *data) {
	node->character = key[0];
	if (!key[1]) { /* (strlen(key) == 1) { */
		node->data = data;
		node->sister = 0;
		node->children = 0;
		return;
	}
	else {
		node->data = 0;
		node->sister = 0;
		node = node->children = (STNode *)malloc(sizeof(STNode));
#ifdef ST_REFC
        sCurrentTable->refc++;
#endif
		ST_Fill(node, &key[1], data);
	}
}

/*
 * Function: ST_AddRec
 * -------------------
 * Internal.
 */
int ST_AddRec(STNode *node, const char *key, void *data) {
	if (node->character == key[0]) {
		if (!key[1]) { /* (strlen(key) == 1) { */
			if (node->data == 0) {
				node->data = data;
				return 1;
			}
			else {
				return 0;
            }
		}
			
		if (node->children) {
			return ST_AddRec(node->children, &key[1], data);
        }
		else {
			node->children = (STNode *)malloc(sizeof(STNode));
#ifdef ST_REFC
            sCurrentTable->refc++;
#endif
			ST_Fill(node->children, &key[1], data);
			return 1;
		}
	}
	else {
		if (node->sister) {
			return ST_AddRec(node->sister, key, data);
        }
		else {
			node->sister = (STNode *)malloc(sizeof(STNode));
#ifdef ST_REFC
            sCurrentTable->refc++;
#endif
			ST_Fill(node->sister, key, data);
			return 1;
		}
	}
}

/*
 * Function: ST_AddData
 * --------------------
 * Add data with an access key to symbol table.
 *
 * st:   symbol table
 * key:  access key
 * data: data to be added
 */
int ST_AddData(STable *st, const char *key, void *data) {
	STNode *node;

#ifdef ST_REFC
    sCurrentTable = st;        
#endif
	
	node = st->rootNode;
	if (node == 0) {
		st->rootNode = node = (STNode *)malloc(sizeof(STNode));
#ifdef ST_REFC
        sCurrentTable->refc++;
#endif
		ST_Fill(node, key, data);
		return 1;
	}
	return ST_AddRec(node, key, data);
}

/*
 * Function: ST_GetDataRec
 * -----------------------
 * Internal.
 */
void *ST_GetDataRec(STNode *node, const char *key) {
	if (node == 0) return 0;
		
    if (!key[1]) { /*if (strlen(key) == 1) {*/
		if (node->character == key[0]) {
			return node->data;
        }
		else {
			return ST_GetDataRec(node->sister, key);
        }
	}
	else {
		if (node->character == key[0]) {
			return ST_GetDataRec(node->children, key + 1);
        }
		else {
			if (node->sister == 0) {
				return 0;
            }
			else {
				return ST_GetDataRec(node->sister, key);
            }
		}
	}
}

/*
 * Function: ST_FindNodeRec
 * ------------------------
 * Internal.
 */
STNode *ST_FindNodeRec(STNode *node, const char *key) {
	if (node == 0) return 0;
		
	if (!key[1]) { /*(strlen(key) == 1) {*/
		if (node->character == key[0]) {
			return node;
        }
		else {
			return ST_FindNodeRec(node->sister, key);
        }
	}
	else {
		if (node->character == key[0]) {
			return ST_FindNodeRec(node->children, key + 1);
        }
		else {

			if (node->sister == 0) {
				return 0;
            }
			else {
				return ST_FindNodeRec(node->sister, key);
            }
		}
	}
}

/*
 * Function: ST_GetData
 * --------------------
 * Return data for access key from symbol table if the key exists.
 *
 * st:  symbol table
 * key: access key
 *
 * Returns: data or 0
 */
void *ST_GetDataOld(STable *st, const char *key) {
    void *data = 0;
	
	data = ST_GetDataRec(st->rootNode, key);

    return data;
}

/*
 * Function: ST_GetData
 * --------------------
 * Return data for access key from symbol table if the key exists.
 *
 * st:  symbol table
 * key: access key
 *
 * Returns: data or 0
 */
void *ST_GetData(STable *st, const char *key) {
    STNode *node = st->rootNode;
    if (!key|| !*key) return 0;
    do {
        if (!node) return 0;
        if (node->character == *key) {
            key++;
            if (!*key) return node->data;
            else node = node->children;
        }
        else {
            node = node->sister;
        }
    } while(1);
}

/*
 * Function: ST_FindNode
 * ---------------------
 * Internal.
 */
STNode *ST_FindNodeOld(STable *st, const char *key) {
    STNode *node = 0;
    node = ST_FindNodeRec(st->rootNode, key);
    return node;
}

/*
 * Function: ST_FindNode
 * ---------------------
 * Internal.
 */
STNode *ST_FindNode(STable *st, const char *key) {
    STNode *node = st->rootNode;
    if (!key|| !*key) return 0;
    do {
        if (!node) return 0;
        if (node->character == *key) {
            key++;
            if (!*key) return node;
            else node = node->children;
        }
        else {
            node = node->sister;
        }
    } while(1);
}


/*
 * Function: ST_DeleteSubKeys
 * ------------------------
 * Delete all sub keys in symbol table. For a parent key "foo", all keys
 * beginning with "foo" will be deleted, but "foo" itself will not be.
 *
 * st:  symbol table
 * key: parent access key
 */
void ST_DeleteSubKeys(STable *st, const char *key) {
    STNode *node = ST_FindNode(st, key);
#ifdef ST_REFC
    sCurrentTable = st;
#endif
    if (node) {
        ST_ClearRec(node->children, st->freeData);
        node->children = 0;
    }
}

/*
 * Function: ST_DeleteKeyRec
 * -------------------------
 * Internal.
 */
int ST_DeleteKeyRec(STNode *node, const char *key, int deleteSubKeys, void (*freeData)(void *)) {
	if (node == 0) return 0;
		
	if (!key[1]) { /* strlen(key) == 1) { */
		if (node->character == key[0]) {
            if (node->data) {
                if (freeData) freeData(node->data);
                node->data = 0;
            }
            if (deleteSubKeys) {
                ST_ClearRec(node->children, freeData);
                node->children = 0;
    			return 1;
            }
            else {
                if (node->children) {
                    return 0;
                }
                else {
                    return 1;
                }
            }
        }
		else {
            if (ST_DeleteKeyRec(node->sister, key, deleteSubKeys, freeData)) {
                STNode *nextSister = node->sister->sister;
                free(node->sister);
#ifdef ST_REFC
                sCurrentTable->refc--;
#endif
                node->sister = nextSister;
                if (node->data || node->children) {
                    return 0;
                }
                else {
                    return 1;
                }
            }
            else {
                return 0;
            }
        }
	}
	else {
		if (node->character == key[0]) {
			if (ST_DeleteKeyRec(node->children, &key[1], deleteSubKeys, freeData)) {
                STNode *childSister = node->children->sister;
                free(node->children);
#ifdef ST_REFC
                sCurrentTable->refc--;
#endif
                node->children = childSister;
                if (node->data || node->children) {
                    return 0;
                }
                else {
                    return 1;
                }
            }
            else {
                return 0;
            }
        }
		else {
			if (node->sister == 0) {
				return 0;
            }
			else {
				if (ST_DeleteKeyRec(node->sister, key, deleteSubKeys, freeData)) {
                    STNode *nextSister = node->sister->sister;
                    free(node->sister);
#ifdef ST_REFC
                    sCurrentTable->refc--;
#endif
                    node->sister = nextSister;
                    if (node->data || node->children) {
                        return 0;
                    }
                    else {
                        return 1;
                    }
                }
                else {
                    return 0;
                }
            }
		}
	}
}

/*
 * Function: ST_DeleteKey
 * ----------------------
 * Delete a key.
 *
 * st:  symbol table
 * key: access key
 */
void ST_DeleteKey(STable *st, const char *key) {
#ifdef ST_REFC
    sCurrentTable = st;
#endif
    if (ST_DeleteKeyRec(st->rootNode, key, 0, st->freeData)) {
        STNode *sister = st->rootNode->sister;
        free(st->rootNode);
#ifdef ST_REFC
        sCurrentTable->refc--;
#endif
        st->rootNode = sister;
    }
}

/*
 * Function: ST_DeleteKeyAndSubKeys
 * --------------------------------
 * Delete a key and its sub keys. For a key "foo", all keys beginning with "foo"
 * will be deleted.
 *
 * st:  symbol table
 * key: access key
 */
void ST_DeleteKeyAndSubKeys(STable *st, const char *key) {
#ifdef ST_REFC
    sCurrentTable = st;
#endif
    if (ST_DeleteKeyRec(st->rootNode, key, 1, st->freeData)) {
        STNode *sister = st->rootNode->sister;
        free(st->rootNode);
#ifdef ST_REFC
        sCurrentTable->refc--;
#endif
        st->rootNode = sister;
    }
}

/*
 * Function: ST_DuplicateRec
 * -------------------------
 * Internal.
 */
void ST_DuplicateRec(STNode *dst, STNode *src, void *(*copyData)(void *)) {
    if (!src) return;
    dst->character = src->character;
    dst->children = 0;
    dst->sister = 0;
    dst->data = copyData(src->data);
    if (src->children) {
        dst->children = (STNode *)malloc(sizeof(STNode));
#ifdef ST_REFC
        sCurrentTable->refc++;
#endif
        ST_DuplicateRec(dst->children, src->children, copyData);
    }
    if (src->sister) {
        dst->sister = (STNode *)malloc(sizeof(STNode));
#ifdef ST_REFC
        sCurrentTable->refc++;
#endif
        ST_DuplicateRec(dst->sister, src->sister, copyData);
    }
}

/*
 * Function: ST_DuplicateKeyAndChildren
 * ------------------------------------
 * Duplicate data of key and all subkeys and their data to another key.
 *
 * st: symbol table
 * dst: destination key
 * key: source key
 */
void ST_DuplicateKeyAndChildren(STable *st, const char *dst, const char *key) {
    STNode *srcNode;
    STNode *dstNode;
    srcNode = ST_FindNode(st, key);
    if (!srcNode) return;
    ST_AddData(st, dst, st->copyData(srcNode->data));
    dstNode = ST_FindNode(st, dst);
    dstNode->children = (STNode *)malloc(sizeof(STNode));
#ifdef ST_REFC
    sCurrentTable = st;
    st->refc++;
#endif
    if (!srcNode->children) return;
    ST_DuplicateRec(dstNode->children, srcNode->children, st->copyData);
}

/*
 * Function: ST_PrintNode
 * ----------------------
 * Internal.
 */
void ST_PrintNode(STNode *node, int depth) {
	int i;
	
	if (node == 0)
		return;
		
	for (i = 0; i < depth; i++) printf(" ");
	printf("%c", node->character);
	if (node->data) {
		printf("*\n");
    }
	else {
		printf("\n");
    }
	ST_PrintNode(node->children, depth + 1);		
	ST_PrintNode(node->sister, depth);
}

/*
 * Function: ST_Print
 * ------------------
 * Print table.
 *
 * st: symbol table
 */
void ST_Print(STable *st) {
	ST_PrintNode(st->rootNode, 0);
}

