/*
 * File: symbol_table.h
 * --------------------
 * Handles tree structures for storing data through access keys.
 *
 * By: Marcus 2018
 */

#ifndef __SYMBOL_TABLE_H__ST_GetData
#define __SYMBOL_TABLE_H__

/* Comment out to remove node reference count and asserts. */
#define ST_REFC

/*
 * Struct: STNode
 * --------------
 * Internal.
 */
typedef struct sSTNode STNode;
struct sSTNode {
	STNode *sister;
	STNode *children;
	char character;
	void *data;
};

/*
 * Struct: STable
 * --------------
 * Symbol table.
 */
typedef struct {
	STNode *rootNode;
	int keyCount;
    void (*freeData)(void *data);
    void *(*copyData)(void *data);
#ifdef ST_REFC
    int refc;
#endif
} STable;

/*
 * Function: ST_Init
 * -----------------
 * Return a new symbol table. Use ST_Clear before freeing the returned symbol
 * table.
 *
 * freeDataFunc: function to be called for destroying data added to the table
 *               with ST_AddData. Any time data is removed from the table, this
 *               function will be called.
 *
 * Returns: symbol table
 */
STable *ST_Init(void (*freeDataFunc)(void *), void *(*copyDataFunc)(void *));

/*
 * Function: ST_Clear
 * ------------------
 * Clear symbol table. You need to manually free the table itself.
 *
 * st: symbol table
 */
void ST_Clear(STable *st);

/*
 * Function: ST_AddData
 * --------------------
 * Add data with an access key to symbol table.
 *
 * st:   symbol table
 * key:  access key
 * data: data to be added
 */
int ST_AddData(STable *st, const char *key, void *data);

/*
 * Function: ST_GetData
 * --------------------
 * Return data for access key from symbol table if the key exists.
 *
 * st:  symbol table
 * key: access key
 *
 * Returns: data or 0
 */
void *ST_GetData(STable *st, const char *key);

/*
 * Function: ST_DeleteKey
 * ----------------------
 * Delete a key.
 *
 * st:  symbol table
 * key: access key
 */
void ST_DeleteKey(STable *st, const char *key);

/*
 * Function: ST_DeleteKeyAndSubKeys
 * --------------------------------
 * Delete a key and its sub keys. For a key "foo", all keys beginning with "foo"
 * will be deleted.
 *
 * st:  symbol table
 * key: access key
 */
void ST_DeleteKeyAndSubKeys(STable *st, const char *key);

/*
 * Function: ST_DeleteSubKeys
 * ------------------------
 * Delete all sub keys in symbol table. For a parent key "foo", all keys
 * beginning with "foo" will be deleted, but "foo" itself will not be.
 *
 * st:  symbol table
 * key: parent access key
 */
void ST_DeleteSubKeys(STable *st, const char *key);

/*
 * Function: ST_DuplicateKeyAndChildren
 * ------------------------------------
 * Duplicate data of key and all subkeys and their data to another key.
 *
 * st: symbol table
 * dst: destination key
 * key: source key
 */
void ST_DuplicateKeyAndChildren(STable *st, const char *dst, const char *key);

/*
 * Function: ST_FindNode
 * ---------------------
 * Internal.
 */
STNode *ST_FindNode(STable *st, const char *key);

/*
 * Function: ST_Print
 * ------------------
 * Print table.
 *
 * st: symbol table
 */
void ST_Print(STable *st);

#endif

